"""
    test_collada
    ~~~~~~~~~~~~
"""

import unittest

from domonic.xml.collada import *

# import requests
# from mock import patch

# from domonic import domonic

# from domonic.decorators import silence


class TestCase(unittest.TestCase):

    # @silence
    def test_domonic_collada(self):
        col = COLLADA()
        print(col)


if __name__ == '__main__':
    unittest.main()



"""
    test_sitemap
    ~~~~~~~~~~~~
"""

import unittest
from datetime import datetime

from domonic import domonic
from domonic.decorators import silence
from domonic.xml.sitemap import *

# import requests
# from mock import patch




class TestCase(unittest.TestCase):

    # @silence
    def test_sitemap(self):

        # a sitemap index contains a list of sitemaps .i.e
        doc = sitemapindex(
            sitemap(
                loc('https://x.net/egypt/post-sitemap.xml'),
                lastmod('2021-07-08T13:12:16+00:00')
            )
        )

        print(doc)
        print(str(doc))

        doc = sitemap(
            url(
                loc('https://xyz.net'),
                lastmod('2021-07-08T13:12:16+00:00'),  # pass a date as string. if no data is passed the current date is used
                changefreq('weekly'),
                priority(0.5)
            )
        )

        print(doc)
        print(str(doc))

    def test_sitemapindex(self):

        from domonic.xml.sitemap import sitemap, sitemapindex, url, urlset

        # sm = sitemapindex()
        # sm.addChild(sitemap(loc('https://abd.net/sitemap1.xml'), lastmod(datetime.datetime.now())))
        # sm.addChild(sitemap(loc('https://abd.net/sitemap2.xml'), lastmod(datetime.datetime.now())))
        # sm.addChild(sitemap(loc('https://abd.net/sitemap3.xml'), lastmod(datetime.datetime.now())))
        # print(f"{sm!s}")
        # print(f"{sm!r}")

        sm = sitemapindex()
        sm += sitemap(loc('https://abd.net/sitemap1.xml'), lastmod(str(datetime.datetime.now())))
        sm += sitemap(loc('https://abd.net/sitemap2.xml'), lastmod(str(datetime.datetime.now())))
        sm += sitemap(loc('https://abd.net/sitemap3.xml'), lastmod(str(datetime.datetime.now())))

        print(f"{sm!s}")
        print(f"{sm!r}")
        # print(f"{sm!a}")
        print(f"{sm}")

    @silence
    def test_loadsitemap(self):
        from domonic.xml.sitemap import get_sitemap

        # sm = get_sitemap('https://x.net/sitemap.xml')
        sm = get_sitemap('https://x.net/merchants/ar/sitemap_index.xml')
        print(sm)

    @silence
    def test_parse_sitemapindex(self):
        # pass
        # https://x.net/merchants/ar/page-sitemap.xml
        from domonic.xml.sitemap import get_sitemap
        sm = get_sitemap('https://x.net/merchants/ar/page-sitemap.xml')
        print(sm)
        # pass

    @silence
    def test_element_class(self):
        # could the existing framework have been used. i.e.
        # from domonic.dom import Document
        # sitemapindex = Document.createElement('sitemapindex') # TODO - should createElementClass be a method of domonic Document?
        pass


if __name__ == '__main__':
    unittest.main()



"""
    test_javascript
    ~~~~~~~~~~~~~~~
    unit tests for domonic.javascript
"""

import time
import unittest
from inspect import stack
from unittest.mock import Mock

from domonic.javascript import *
from domonic.javascript import (URL, Array, Date, Global, Math, Object, String,
                                Window)

# import requests
# from mock import patch






class TestCase(unittest.TestCase):

    # domonic.javascript.Math

    def test_object(self):

        o = Object()
        print(o)
        print(type(o))

        myObj = Object()
        string = 'myString'
        rand = Math.random()
        obj1 = Object()
        myObj.type = 'Dot syntax'
        myObj['date created'] = 'String with space'
        myObj[string] = 'String value'
        myObj[rand] = 'Random Number'
        myObj[obj1] = 'Object'
        myObj[''] = 'Even an empty string'

        assert myObj.type == 'Dot syntax'
        assert myObj['date created'] == 'String with space'
        assert myObj[string] == 'String value'
        assert myObj[rand] == 'Random Number'
        # assert myObj[obj1] == 'Object' # TODO - does js do this?
        assert myObj[''] == 'Even an empty string'

        assert o is not myObj

        myCar = Object()
        propertyName = 'make'
        myCar[propertyName] = 'Ford'
        assert myCar[propertyName] == 'Ford'
        propertyName = 'model'
        myCar[propertyName] = 'Mustang'
        assert myCar[propertyName] == 'Mustang'

        def showProps(obj, objName):
            result = ''
            for i in obj:
                if obj.hasOwnProperty(i):
                    result += objName + "." + str(i) + "= " + str(obj[i]) + "\n"
            return result

        showProps(myCar, "myCar")
        # print(showProps(myCar, "myCar"))

        obj = {'a': 1}
        copy = Object.assign({}, obj)
        assert copy == {'a': 1}

        # print(Object().fromEntries())
        arr = [['0', 'a'], ['1', 'b'], ['2', 'c']]
        obj = Object.fromEntries(arr)
        print(obj)
        assert obj == {'0': "a", '1': "b", '2': "c"}

        obj = {'foo': 'bar', 'baz': 42}
        print(Object.entries(obj))
        assert Object.entries(obj) == [['foo', 'bar'], ['baz', 42]]

        # array like object
        obj = {'0': 'a', '1': 'b', '2': 'c'}
        assert Object.entries(obj) == [['0', 'a'], ['1', 'b'], ['2', 'c']]

        # def listAllProperties(o):
        #     result = []
        #     objectToInspect = o
        #     while objectToInspect != None:
        #         print(objectToInspect)
        #         objectToInspect = Object.getPrototypeOf(objectToInspect)
        #         result = Array(result).concat(Object.getOwnPropertyNames(objectToInspect))

        #     return result
        # print(listAllProperties(myCar))


        # array like object with random key ordering
        # anObj = {'100': 'a', '2': 'b', '7': 'c'}
        # print(anObj)
        # print(Object.entries(anObj))
        # assert Object.entries(anObj) == [['2', 'b'], ['7', 'c'], ['100', 'a']]

        # returns an empty array for any primitive type
        assert Object.entries(100) == []

        # iterate through key-value gracefully
        obj = {'a': 5, 'b': 7, 'c': 9}
        for key, value in Object.entries(obj):
            print(f'{key} {value}')  # "a 5", "b 7", "c 9"

        # class Car(Object):
        #     def __init__(self, make, model, year):
        #         super().__init__()
        #         self.make = make
        #         self.model = model
        #         self.year = year
        #         # super().__init__()

        # mycar = Car('Eagle', 'Talon TSi', 1993)
        # print(mycar)
        # print(mycar.make)
        # print(mycar.__attribs__)

    # Animal properties and method encapsulation


    # TODO - to get reference back to self. in a dict it needs to readd the method and pass self
    Animal = {
        'type': 'Invertebrates',  # Default value of properties
        # 'displayType': lambda self: print("STACK!!!!!",stack()[1].function)  # Method which will display type of Animal
        'displayType': lambda self: print(self.type)
    }
    animal1 = Object.create(Animal)
    print(animal1)
    # print(animal1['type'])
    print(animal1.__dict__)
    animal1.displayType(animal1)  # Output:Invertebrates #TODO - need to work without passing self

    fish = Object.create(Animal)
    fish.type = 'Fishes'
    fish.displayType(animal1)  # Output:Fishes



    def test_domonic_abs(self):
        # python -m unittest tests.test_javascript.TestCase.test_domonic_abs

        self.assertEqual(Math.abs('-1'), 1)
        self.assertEqual(Math.abs(-2), 2)
        self.assertEqual(Math.abs(None), 0)
        self.assertEqual(Math.abs(''), 0)
        self.assertEqual(Math.abs([]), 0)
        self.assertEqual(Math.abs([2]), 2)
        self.assertEqual(Math.abs([1, 2]), None)
        self.assertEqual(Math.abs({}), None)
        self.assertEqual(Math.abs('string'), None)
        self.assertEqual(Math.abs(), None)

        self.assertEqual(100, Math.abs(-100.0))

    def test_domonic_LN2(self):
        print("test_domonic_LN2:::")
        print(Math.LN2)

    def test_domonic_LOG2E(self):
        print("test_domonic_LOG2E:::")
        print(Math.LOG2E)

    def test_domonic_LOG10E(self):
        print("test_domonic_LOG10E:::")
        print(Math.LOG10E)

    def test_domonic_PI(self):
        print("test_domonic_PI:::")
        print(Math.PI)

    def test_domonic_SQRT1_2(self):
        print("test_domonic_SQRT1_2:::")
        print(Math.SQRT1_2)

    def test_domonic_SQRT2(self):
        print("test_domonic_SQRT2:::")
        print(Math.SQRT2)

    def test_domonic_acos(self):
        print("test_domonic_acos:::")
        # print( Math.acos(-100) ) # TODO - fails numbers greater than 1 or lower than -1 - raise Error?
        print(Math.acos(0.5))

    def test_domonic_acosh(self):
        print("test_domonic_acosh:::")
        # print( Math.acosh(-100) ) # TODO - fails under zero - rause error?
        print(Math.acosh(100))

    def test_domonic_asin(self):
        print("test_domonic_asin:::")
        # print( Math.asin(-100) ) # TODO - fails numbers greater than 1 or lower than -1 - raise Error?
        print(Math.asin(0.5))

    def test_domonic_asinh(self):
        print("test_domonic_asinh:::")
        print(Math.asinh(-100))

    def test_domonic_atan(self):
        print("test_domonic_atan:::")
        print(Math.atan(-100))

    def test_domonic_atan2(self):
        print("test_domonic_atan2:::")
        print(Math.atan2(-100, 100))

    def test_domonic_atanh(self):
        print("test_domonic_atanh:::")
        # print( Math.atanh(-100) ) # TODO - fails numbers greater than 1 or lower than -1 - raise Error?
        print(Math.atanh(0.5))

    def test_domonic_cbrt(self):
        print("test_domonic_cbrt:::")
        # print( Math.cbrt(-100) ) # TODO - fails on negative numbers - raise Error?
        print(Math.cbrt(100))

    def test_domonic_ceil(self):
        print("test_domonic_ceil:::")
        print(Math.ceil(-100))

    def test_domonic_cos(self):
        print("test_domonic_cos:::")
        print(Math.cos(-100))

    def test_domonic_cosh(self):
        print("test_domonic_cosh:::")
        print(Math.cosh(-100))

    def test_domonic_E(self):
        self.assertEqual(2.718281828459045, Math.E)

    def test_domonic_exp(self):
        print("test_domonic_exp:::")
        print(Math.exp(-100))

    def test_domonic_floor(self):
        print("test_domonic_floor:::")
        print(Math.floor(-100))

    def test_domonic_LN10(self):
        self.assertEqual(2.302585092994046, Math.LN10)

    def test_domonic_log(self):
        print("test_domonic_log:::")
        # print( Math.log(-100,100) ) # TODO - fails on negative numbers - raise Error?
        print(Math.log(100, 10))

    def test_domonic_max(self):
        print("test_domonic_max:::")
        print(Math.max(-100, 100))

    def test_domonic_min(self):
        print("test_domonic_min:::")
        print(Math.min(-100, 100))

    def test_domonic_random(self):
        print("test_domonic_random:::")
        print(Math.random())

    def test_domonic_round(self):
        print("test_domonic_round:::")
        print(Math.round(-100))

    def test_domonic_pow(self):
        print("test_domonic_pow:::")
        print(Math.pow(100, 10))

    def test_domonic_sin(self):
        print("test_domonic_sin:::")
        print(Math.sin(-100))

    def test_domonic_sinh(self):
        print("test_domonic_sinh:::")
        print(Math.sinh(-100))

    def test_domonic_sqrt(self):
        print("test_domonic_sqrt:::")
        # print( Math.sqrt(-100) ) # TODO - fails on negative numbers - raise Error? check js behaviour
        print(Math.sqrt(100))

    def test_domonic_tan(self):
        print("test_domonic_tan:::")
        print(Math.tan(-100))

    def test_domonic_tanh(self):
        # print("test_domonic_tanh:::")
        # print(Math.tanh(-100))
        assert Math.tanh(0) == 0
        assert Math.tanh(1) == 0.7615941559557649
        assert Math.tanh(2) == 0.9640275800758169
        assert Math.tanh(3) == 0.9950547536867305

    def test_domonic_trunc(self):
        print("test_domonic_trunc:::")
        print(Math.trunc(-100))

    # def test_domonic_math_test(self):
    #   print("test_domonic_math_test:::")
    #   print( Math.abs(-100)*Math.random()*10 )

    # domonic.javascript.Global

    def test_domonic_isNaN(self):
        self.assertEqual(True, Global.isNaN("yo"))
        self.assertEqual(False, Global.isNaN(1))

    def test_domonic_Number(self):
        self.assertEqual(1, Global.Number(1))
        self.assertEqual("NaN", Global.Number("test"))
        self.assertEqual(2, Global.Number("1") + Global.Number("1.0"))

    def test_domonic_window_console_log(self):
        # window = Window()
        # Window().console.log("test this")
        # window.console.log("test this")

        # c = Console()
        # c.log()
        # Console.log('test')
        pass

    def test_domonic_window_alert(self):
        # Window().alert("test this 2")
        window = Window()
        window.alert("test this 2")

    def test_domonic_window_document_baseURI(self):
        # Window().alert("test this 2")
        # window = Window()
        # window.alert("test this 2")
        # print(window.document.baseURI)
        # window.document.baseURI = "eventual.technology"
        # print("=",window.document.baseURI)

        pass

    '''
    def test_domonic_window_location(self):
        # Window().alert("test this 2")
        window = Window()
        # window.alert("test this 2")
        print("window.location")
        print(window.location)
        window.location = "eventual.technology"
        print("window.location.uri")
        print(window.location)
        print(str(window.location))
        print(window.location.href)
    '''

    def test_domonic_global_encodeURIComponent(self):

        msg = "Test encoding this string! 123 aweseome"
        enc_msg = Global.encodeURIComponent(msg)
        print(enc_msg)
        # print( Global.decodeURIComponent(bytes(enc_msg, encoding="UTF-8")))
        print(Global.decodeURIComponent(enc_msg))

        # Window().alert("test this 2")
        # window = Window()
        # window.alert("test this 2")
        # print(window.document.baseURI)
        # window.document.baseURI = "eventual.technology"
        # print("=",window.document.baseURI)
        pass

    def test_javascript_date(self):
        print("test_javascript_date::::::::::::::::::")
        d = Date()
        print(d.getDate())
        print(d.getDay())
        print(d.getFullYear())
        print(d.getHours())
        print(d.getMilliseconds())
        print(d.getMinutes())
        print(d.getMonth())
        print(d.getSeconds())
        print(d.getTime())
        # print( d.getTimezoneOffset() )
        print(d.getUTCDate())
        print(d.getUTCDay())
        print(d.getUTCFullYear())
        print(d.getUTCHours())
        print(d.getUTCMilliseconds())
        print(d.getUTCMinutes())
        print(d.getUTCMonth())
        print(d.getUTCSeconds())
        print(d.getYear())
        print(d.now())
        # print( d.onstorage() )
        # print( d.ontimeupdate() )
        print(d.parse("July 1981"))
        print(d.setDate(1))
        print(d.setFullYear('1982'))
        print(d.setHours(2))
        # print( d.setItem() )
        print(d.setMilliseconds(12345))
        print(d.setMinutes(10))
        print(d.setMonth(10))
        print(d.setSeconds(10))
        print(d.setTime(1000))
        print(d.setUTCDate(1))
        print(d.setUTCFullYear(1928))
        print(d.setUTCHours(3))
        print(d.setUTCMilliseconds(54321))
        print(d.setUTCMinutes(50))
        print(d.setUTCMonth(3))
        print(d.setUTCSeconds(11))
        print(d.setYear(1987))
        print(d.toDateString())
        print(d.toGMTString())
        print(d.toJSON())
        print(d.toISOString())
        print(d.toLocaleDateString())
        print(d.toLocaleString())
        print(d.toLocaleTimeString())
        print(d.toTimeString())
        print(d.toUTCString())
        print(d.UTC())

        print(Date(1415988000))

    # def test_domonic_parse(self):
        # page = domonic.parse("<html><body>'some content'</body></html>")
        # page = domonic.parse("<html><body></body></html>")
        # print(page)

    # def test_domonic_get(self):
    #     print("test_domonic_get-----------=-----------=-----------=-----------=-----------=-----------=-----------=")
    #     page = eval(domonic.get("http://eventual.technology"))
    #     dir(page)

    # TODO - this was move to webapi. tests are working by proxy
    def test_javascript_url(self):
        url = URL('https://somesite.com/blog/article-one#some-hash')
        # print('TESTS:')
        # print(url)
        assert url.href == 'https://somesite.com/blog/article-one#some-hash'
        assert url.protocol == 'https'
        assert url.host == 'somesite.com'
        assert url.hostname == 'somesite.com'
        # assert url.port == ''  # TODO - check js none or empty
        assert url.pathname == '/blog/article-one'
        assert url.hash == '#some-hash'
        assert url.toString() == 'https://somesite.com/blog/article-one#some-hash'
        # print(url.protocol)
        url.protocol = "http"
        # print(url.protocol)
        assert url.protocol == "http"
        assert url.href == "http://somesite.com/blog/article-one#some-hash"

        url.host = 'test.com'
        assert url.href == "http://test.com/blog/article-one#some-hash"
        assert url.host == "test.com"
        assert url.hostname == "test.com"
        url.port = 8983
        assert url.href == "http://test.com:8983/blog/article-one#some-hash"
        assert url.port == 8983
        # print(url.toString())

    # def test_javascript_window(self):
        # print('asdf')
        # print(window)
        # print(window.location)

        # window.location = "https://google.com"
        # print(window.location.href)
        # pass

    def test_javascript_array(self):
        myarr = Array("1", "2", 3, {"4": "four"}, 5, [6])
        assert isinstance(myarr, Array)
        assert myarr.length == 6
        assert myarr.includes("1")
        assert myarr.includes(3)
        assert not myarr.includes(10)
        assert myarr.indexOf(10) == -1
        assert myarr.indexOf("1") == 0
        assert myarr.indexOf([6]) == 5
        assert myarr[1] == "2"
        assert len(myarr) == 6
        assert myarr == Array("1", "2", 3, {"4": "four"}, 5, [6])
        # print(myarr.join('---'))  #  TODO - test some js ones
        # assert myarr.join('---') == "1---2---3---[object Object]---5---6"
        assert myarr.join('---') == "1---2---3---{'4': 'four'}---5---[6]"
        # print(myarr.lastIndexOf("1"))
        assert myarr.lastIndexOf("1") == 0
        assert myarr.lastIndexOf(3) == 2
        print('fails to assert:', myarr.reverse())  # TODO - not passing but looks right? - also not fixed by equality update
        # assert myarr.reverse() == [[6], 5, {'4': 'four'}, 3, '2', '1']
        myarr = Array([[6], 5, {'4': 'four'}, 3, '2', '1'])
        assert myarr.slice(0, 1) == [[6]]
        assert myarr == Array([[6], 5, {'4': 'four'}, 3, '2', '1'])
        # print(myarr.splice(1))  # TODO - not passing but looks right?
        assert myarr.splice(1) == [5, {'4': 'four'}, 3, '2', '1']
        assert myarr[0][0] == 6
        # tests equality. Array == list
        assert myarr == [[6]]
        # casting back to list
        myarr = list(myarr)
        assert myarr == [[6]]
        # test casting
        myarr = ["1", "a", "b", "c"]
        assert Array(myarr).splice(1, 1) == ['a']
        assert myarr == ["1", "b", "c"]
        myarr = Array(["1", "a", "b", "c"])
        assert myarr.pop() == "c"
        assert myarr == ["1", "a", "b"]
        myarr.push(7)
        assert myarr == ["1", "a", "b", 7]
        assert myarr.unshift('z') == 5
        assert myarr == ["z", "1", "a", "b", 7]
        assert myarr.shift() == "z"
        assert myarr == ["1", "a", "b", 7]
        assert myarr.concat() == ["1", "a", "b", 7]
        # assert myarr.concat(['a', 'b', 'c']) == ["1", "a", "b", 7, "a", "b", "c"]
        assert myarr.concat(['a', 'b', 'c'], ['d', 'e', 'f']) == ["1", "a", "b", 7, "a", "b", "c", "d", "e", "f"]
        # make array do both python and javascript methods.
        myarr = Array("1", "2", 3, {"4": "four"}, 5, [6])
        myarr.append('test')
        myarr = Array([2, 3, 1, 'a', 'b', 5, '2'])
        assert myarr.sort() == [1, 2, '2', 3, 5, 'a', 'b']
        assert myarr.reverse() == ['b', 'a', 5, 3, '2', 2, 1]
        # print(myarr.fill()) # note - js returns list of undefined
        assert myarr.fill() == [None, None, None, None, None, None, None]
        # print(myarr.fill(1))
        assert myarr.fill(1) == [1, 1, 1, 1, 1, 1, 1]
        # print(myarr.fill(1, 1))
        assert myarr.fill(1, 1) == [1, 1, 1, 1, 1, 1, 1]
        # print(myarr.fill(1, 1, 3))  # TODO - more fill tests
        # print(myarr.isArray()) # fails as it should as its a static method.
        assert Array.isArray(myarr) == True

        myarr = Array([3, 4, 2, 'b', 'c', 6, 3])
        # print(myarr.map(lambda x: x + 1 if type(x) == int else chr(ord(x) + 1)))
        assert myarr.map(lambda x: x + 1 if type(x) == int else chr(ord(x) + 1)) == [4, 5, 3, 'c', 'd', 7, 4]
        # print(myarr.filter()) # passing nothing fails like javascript
        with self.assertRaises(TypeError):
            myarr.filter()
        # print(myarr.filter(lambda x: x == "a"))
        assert myarr.filter(lambda x: x == "a") == []
        # print(myarr.filter(lambda x: x == 3))
        assert myarr.filter(lambda x: x == 3) == [3, 3]
        # print(myarr.reduce()) # passing nothing fails like javascript
        with self.assertRaises(TypeError):
            myarr.reduce()
        # print(myarr.reduce(lambda x, y: x + y))
        # print(myarr)
        myarr = Array([3, 4, 2, 6, 3])
        # print(myarr.reduce(lambda x, y: x + y))
        assert myarr.reduce(lambda x, y: x + y) == 18
        # print(myarr)
        assert myarr.reduce(lambda x, y: x + y, 10) == 28
        # print(myarr.reduceRight())
        with self.assertRaises(TypeError):
            myarr.reduceRight()
        # print(myarr.reduceRight(lambda x, y: x + y))
        assert myarr.reduceRight(lambda x, y: x + y) == 18
        # print(myarr)
        assert myarr.reduceRight(lambda x, y: x + y, 10) == 28

        with self.assertRaises(TypeError):
            myarr.every()
        # print(myarr.every(lambda x: x == "a"))
        assert myarr.every(lambda x: x == "a") == False

        # do a test for true
        myarr = Array([3, 3, 3, 3, 3])
        # print(myarr.every(lambda x: x == 3))
        assert myarr.every(lambda x: x == 3) == True

        myarr = Array([3, 4, 2, 'b', 'c', 6, 3])
        with self.assertRaises(TypeError):
            myarr.some()
        # print(myarr.some(lambda x: x == "a"))
        assert myarr.some(lambda x: x == "b") == True

        # print(myarr.find())
        with self.assertRaises(TypeError):
            myarr.find()

        # print(myarr.find(lambda x: x == "a"))
        assert myarr.find(lambda x: x == "a") == None
        # print(myarr.find(lambda x: x == "b"))
        assert myarr.find(lambda x: x == "b") == "b"
        pass

    def test_javascript_interval(self):

        def hi():
            # print('hi')
            spin = ["|", "/", "-", "\\"]
            print('    ' + spin[int(time.time() * 4) % 4], end="\r")

        test = window.setInterval(hi, 1)
        print("2 secs. I'm just going to do some stuff in the background")

        # keep the test open to see if the intervals fire
        time.sleep(2)

        print('running again')
        window.clearInterval(test)
        print('ran')

    def test_javascript_Number(self):
        # print(Number.MAX_VALUE)
        assert Number.MAX_VALUE == 1.7976931348623157e+308

    def test_javascript_fetch(self):

        TEST_DOMAIN = 'http://eventual.technology'
        urls = ['http://google.com', 'http://linkedin.com', 'http://eventual.technology']  # use your own domains

        print('run 1')
        results = window.fetch(TEST_DOMAIN)
        results.then(lambda r: print(r.text))
        print('run 1 FINISHED')

        def somefunc(response):
            print("I'm a callback", response.ok)
            return response

        mydata = window.fetch(TEST_DOMAIN).then(somefunc)
        print(mydata)
        print(mydata.data)
        print(mydata.data.text)

        print('run 1111')
        results = window.fetch_set(urls)
        print(results)
        print(list(results))
        for r in results:
            if r is not None:
                print(r.ok)
                # print(r.text)

        print('run 2')
        results = window.fetch_threaded(urls)
        print(results)
        print(list(results))
        for r in results:
            if r is not None:
                print(r.ok)
                # print(r.text)

        print('run 3')
        results = window.fetch_pooled(urls, timeout=2)
        print(results)
        for r in results:
            if r is not None:
                print(r.ok)
                # print(r.text)

        print('run 4')
        results = window.fetch(urls[0])
        print(results)
        results.then(lambda r: print(r.text) if r is not None else None)

        print('ran ===')
        # return

        # TEST REGULAR

        global _results

        def get_things():
            global _results
            _results = window.fetch(urls[0])
            print('sup::', _results)

        print('BEFORE')
        test = window.setInterval(get_things, 2000)
        print('AFTER')
        print(_results)
        time.sleep(4)
        print('LATER')
        print(_results)

        print('MAKE SURE TO CLEAR INTERVAL AND RESET RESULTS!')
        window.clearInterval(test)
        _results = []

        # TEST - Threaded interval triggering a CPU pool
        def get_things():
            global _results
            _results = window.fetch_pooled(urls)
            print('sup::', _results)

        print('Are you ready')
        test = window.setInterval(get_things, 1000)
        print("wait, where my results?")
        print(_results)
        time.sleep(4)
        print("Ahhh nice")
        print(_results)
        for r in _results:
            print(r.ok)
            # print(r.text)

        window.clearInterval(test)

        # nice 😎

    def test_javascript_promise(self):
        def do_test(resolve, reject):
            global _intID
            _intID = window.setInterval(resolve, 2000, 'amazing!')
            resolve("once!")
        myPromise = Promise(lambda resolve, reject: do_test(resolve, reject))
        myPromise.then(lambda successMessage: str(successMessage))
        time.sleep(3)
        window.clearInterval(_intID)
        myPromise.then(lambda successMessage: print("Yay! " + str(successMessage)))

    def test_javascript_string(self):
        print("test_javascript_string")
        mystr = String("Some String")

        assert(mystr.toLowerCase() == "some string")
        assert(mystr.toUpperCase() == "SOME STRING")

        # print(type(mystr))
        # print(mystr.length)
        assert(mystr.length == 11)

        assert(mystr.repeat(2) == "Some StringSome String")
        # print(mystr)
        # print(mystr)
        # print(mystr)
        assert(mystr.startsWith('S'))
        # assert(mystr.endsWith('g'))

        # print(">>", mystr.substr(1))
        assert(mystr.substr(1) == 'ome String')

        #substring
        # print(mystr)
        # print(mystr.substring(1, 3))
        assert(mystr.substring(1, 3) == 'om')

        # slice
        # print(mystr.slice(1, 3))
        assert(mystr.slice(1, 3) == 'om')

        # test trim
        mystr = String("   Some String   ")
        assert(mystr.trim() == "Some String")

        # charAt
        mystr = String("Some String")
        assert(mystr.charAt(1) == 'o')
        assert(mystr.charAt(5) == 'S')

        # charCodeAt
        assert(mystr.charCodeAt(1) == 111)
        assert(mystr.fromCharCode(111) == 'o')

        # test
        # assert(mystr.test('a') == True)
        # assert(mystr.test('b') == False)

        # replace
        # print(mystr.replace('S', 'X'))
        assert(mystr.replace('S', 'X') == "Xome String")
        assert(mystr.replace(' ', 'X') == "SomeXString")
        assert(mystr.replace('S', 'X') != "Xome Xtring")

        # localeCompare
        # assert(mystr.localeCompare('a', 'b') == -1)
        # assert(mystr.localeCompare('a', 'a') == 0)
        # assert(mystr.localeCompare('a', 'A') == 1)
        # assert(mystr.localeCompare('a', 'aa') == -1)
        # assert(mystr.localeCompare('a', 'Aa') == -1)

        # search
        mystr = String("Some String")
        assert(mystr.search('a') == False)
        assert(mystr.search('o') == True)

        # substr
        print(mystr.substr(1, 2))
        assert(mystr.substr(1, 2) == 'om')
        assert(mystr.substr(1, 3) == 'ome')
        assert(mystr.substr(1, 4) == 'ome ')
        assert(mystr.substr(1, 5) == 'ome S')


        # toLocaleLowerCase
        # print(mystr.toLocaleLowerCase())
        assert(mystr.toLocaleLowerCase() == 'some string')
        # print(mystr.toLocaleLowerCase())
        assert(mystr.toLocaleLowerCase() == 'some string')

        # toLocaleUpperCase
        # print(mystr.toLocaleUpperCase())
        assert(mystr.toLocaleUpperCase() == 'SOME STRING')

        # compile
        # print(mystr.compile())
        # assert(mystr.compile() == '"Some String"')

        # lastIndex
        # print(mystr.lastIndexOf('o'))
        assert(mystr.lastIndexOf('o') == 1)

        # replace
        assert mystr.codePointAt(1) == 111
        # print(mystr.padEnd(2))
        # print(f"-{mystr}-")
        print(f"---{mystr.padEnd(13)}-")
        assert mystr.padEnd(13) == "Some String  "
        assert mystr.padStart(13) == "  Some String"
        assert mystr.padStart(13, '-') == "--Some String"
        # assert mystr.localeCompare('a', 'a') == 0

        assert mystr.includes('a') == False
        assert mystr.includes('Some') == True
        # assert mystr.matchAll(['a', 'b']) == False # TODO - dont think this is supposed to take lists?
        # assert mystr.match('a', 'b') == False # TODO
        # assert mystr.trimStart(1) == "Some" # TODO
        # assert mystr.trimEnd(1) == "String" # TODO


    def test_javascript_URLSearchParams(self):
        print("test_javascript_URLSearchParams")

        paramsString = "q=test&topic=api"
        searchParams = URLSearchParams(paramsString)

        # Iterate the search parameters.
        for p in searchParams:
            print(p)

        assert searchParams.has("topic") == True  # True
        # print( searchParams.get("topic") )
        assert searchParams.get("topic") == "api"  # True
        # searchParams.getAll("topic"); # ["api"]
        assert searchParams.get("foo") is None  # true
        print(searchParams.toString())
        searchParams.append("topic", "webdev")
        print(searchParams.toString())
        assert searchParams.toString() == "q=test&topic=api&topic=webdev"
        searchParams.set("topic", "More webdev")
        assert searchParams.toString() == "q=test&topic=More+webdev"
        searchParams.delete("topic")
        assert searchParams.toString() == "q=test"

        # GOTCHAS

        paramsString1 = "http://example.com/search?query=%40"
        searchParams1 = URLSearchParams(paramsString1)

        assert searchParams1.has("query") == False
        assert searchParams1.has("http://example.com/search?query") == True

        assert searchParams1.get("query") == None
        searchParams1.get("http://example.com/search?query")  # "@" (equivalent to decodeURIComponent('%40'))

        paramsString2 = "?query=value"
        searchParams2 = URLSearchParams(paramsString2)
        print(searchParams2)
        assert searchParams2.has("query") == True

        url = URL("http://example.com/search?query=%40")

        searchParams3 = URLSearchParams(url.search)

        print(searchParams3)
        # print(str(searchParams3))
        # assert searchParams3.has("query") == True

        base64 = window.btoa(String.fromCharCode(19, 224, 23, 64, 31, 128))  # base64 is "E+AXQB+A"
        print(base64)
        searchParams = URLSearchParams("q=foo&bin=" + str(base64))  # q=foo&bin=E+AXQB+A
        # getBin = searchParams.get("bin")  # "E AXQB A" + char is replaced by spaces
        # print(getBin)
        # window.btoa(window.atob(getBin))  # "EAXQBA==" no error thrown
        # window.btoa(String.fromCharCode(16, 5, 208, 4))  # "EAXQBA==" decodes to wrong binary value
        # getBin.replace(r'/ /g', "+")  # "E+AXQB+A" is one solution

        # or use set to add the parameter, but this increases the query string length
        # searchParams.set("bin2", base64)  # "q=foo&bin=E+AXQB+A&bin2=E%2BAXQB%2BA" encodes + as %2B
        # searchParams.get("bin2")  # "E+AXQB+A"

    def test_javascript_FormData(self):
        print("test_javascript_FormData")
        # f = form(input(_type="text", _name="test", _id="test"))
        # d = FormData(f)
        # print(d)
        pass

    def test_javascript_Worker(self):
        print("test_javascript_Worker")
        # myWorker = Worker('/worker.py')
        # first = document.querySelector('input#number1')
        # second = document.querySelector('input#number2')
        # first.onchange = lambda evt : \
        #     myWorker.postMessage([first.value, second.value])
        #     print('Message posted to worker')
        pass

    def test_javascript_at(self):
        print("test_javascript_at")
        myarr = Array(['a', 'b', 'c', 'd'])
        assert myarr.at(-1) == 'd'
        myarr = ['a', 'b', 'c', 'd']
        myarr = Array(myarr)
        assert myarr.at(-1) == 'd'
        myarr = Array('a', 'b', 'c', 'd')
        assert myarr.at(-1) == 'd'


    # def test_javascript_Node(self):
        # url = require('url');
        # console.log(url.domainToASCII('español.com'))
        # console.log(url.domainToASCII('??.com'))
        # console.log(url.domainToASCII('xn--iñvalid.com'))
        # console.log(url.domainToUnicode('español.com'))
        # console.log(url.domainToUnicode('??.com'))
        # console.log(url.domainToUnicode('xn--iñvalid.com'))


    # def test_javascript_call(self):

    #     class Product():
    #         def __init__(self, name, price):
    #             self.name = name
    #             self.price = price

    #     class Food():
    #         def __init__(self, name, price):
    #             Function(Product).call(self, name, price)
    #             self.category = 'food'

    #     class Toy():
    #         def __init__(self, name, price):
    #             Function(Product).call(self, name, price)
    #             self.category = 'toy'

    #     cheese = Food('feta', 5)
    #     fun = Toy('robot', 40)

    #     print(cheese)
    #     print(fun)

    def test_javascript_called(self):

        from domonic.decorators import called
        from domonic.dQuery import dQuery_el, º

        # import time
        # @called(lambda:time.sleep(2))
        # def anon( data=None ):
        #     print("func you")
        #     print(data)
        # import time
        # @called(time.sleep(2)) # calls right away without lambda. but doesnt pass data. can i detect it?
        # def anon( data=None ):
        #     print("func you")
        #     print(data)

        @called(
            lambda: º.ajax('https://www.google.com'),
            lambda err: print('error:', err))
        def success(data=None):
            print("sweet!")
            print(data.text)

        from domonic.decorators import iife

        @iife()
        def sup():
            print("sup!")
            return True


    def test_javascript_numbersandstrings(self):
        print("test_javascript_numbersnstrings")

        n = Number(1)
        n2 = Number(2)
        print(n + n2)

        s = String('a')
        s2 = String('b')
        print(s + s2)
        print(s * n2)

        test = String("test")
        # print(test - 2) # considering allowing this

        print(test[0:1])

        print(test.toUpperCase())
        print(test.toLowerCase())
        print(test.toLocaleLowerCase())
        print(test.toLocaleUpperCase())


    def test_set(self):

        mySet1 = Set()

        mySet1.add(1)           # Set [ 1 ]
        assert mySet1.size == 1
        assert mySet1.contains(1) == True
        assert mySet1.contains(2) == False

        mySet1.add(5)           # Set [ 1, 5 ]
        assert mySet1.size == 2
        assert mySet1.contains(1) == True
        assert mySet1.contains(2) == False
        assert mySet1.contains(5) == True

        mySet1.add(5)           # Set [ 1, 5 ]
        assert mySet1.size == 2
        assert mySet1.contains(1) == True

        mySet1.add('some text') # Set [ 1, 5, 'some text' ]
        assert mySet1.size == 3
        assert mySet1.contains(1) == True
        assert mySet1.contains(2) == False
        assert mySet1.contains(5) == True
        assert mySet1.contains('some text') == True
        assert mySet1.contains('text') == False

        '''
        # TODO - make the following work. js sets can have dictionaries in them
        o = {'a': 1, 'b': 2}
        mySet1.add(o)  # TODO - ok. so we learned something. i nearly never use sets so this is a nice example of difference between python and javascript sets.
        assert mySet1.size == 4
        assert mySet1.contains(a) == True

        mySet1.add({a: 1, b: 2})   # o is referencing a different object, so this is okay

        assert mySet1.has(1) == True
        assert mySet1.has(3) == False
        assert mySet1.has(5) == True
        assert mySet1.has(Math.sqrt(25)) == True
        assert mySet1.has('Some Text'.toLowerCase()) == True
        assert mySet1.has(o) == True

        mySet1.size == 5

        mySet1.delete(5)    # removes 5 from the set
        assert mySet1.has(5) == False

        mySet1.size == 4

        console.log(mySet1)
        # logs Set(4) [ 1, "some text", {…}, {…} ] in Firefox
        # logs Set(4) { 1, "some text", {…}, {…} } in Chrome
        '''

    def test_setTimeout(self):
        """ Test the Global.setTimeout function calls the callback. """

        callback = Mock()

        args = ("hello", "world")
        kwargs = {"foo": "bar"}

        timer_id = Global.setTimeout(callback, 1000, *args, **kwargs)
        callback.assert_not_called()

        assert timer_id in Global._Global__timers

        time.sleep(1.5)

        callback.assert_called_once()
        callback.assert_called_with(*args, **kwargs)

    def test_clearTimeout(self):
        """ Test that Global.clearTimeout function can cancel a timeout. """

        callback = Mock()

        args = ("hello", "world")
        kwargs = {"foo": "bar"}

        timer_id = Global.setTimeout(callback, 1000, *args, **kwargs)
        callback.assert_not_called()

        assert timer_id in Global._Global__timers
        Global.clearTimeout(timer_id)
        assert timer_id not in Global._Global__timers

        time.sleep(1.5)
        callback.assert_not_called()

    def test_timeouts(self):
        def somefunc():
            print("hi!")
        someID = Global.setTimeout(somefunc, 2000)
        Global.clearTimeout(someID)
        time.sleep(2.5)
        # nice!

        # from domonic.javascript import setTimeout, clearTimeout
        # setTimeout(somefunc, 2000)
        # time.sleep(2.5)

        # from domonic.javascript import setInterval, clearInterval
        # def somefunc2():
            # print("testing interval")
        # interval_id = setInterval(somefunc2, 2000)
        # time.sleep(10)
        # clearInterval(interval_id)
        # time.sleep(4)
        # print('end')

        # from domonic.javascript import window
        # intID = window.setInterval(somefunc2, 2000)
        # time.sleep(10)
        # clearInterval(intID)  # not clearing brings it back to life!

    def test_TypedArrays(self):
        # typed arrays
        # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Typed_arrays
        # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/TypedArray
        # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/ArrayBuffer
        # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/DataView
        # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Int8Array

        b = ArrayBuffer(8)
        # print(b.byteLength)
        assert b.byteLength == 8

        # var buffer = new ArrayBuffer(8);
        view = Int8Array(b)
        print(view)

        # From a length
        int8 = Int8Array(2)
        # print(int8[0])
        # print(Int8Array(2)[0])
        # print(Int8Array(25)[0])
        # print(Int8Array(12)[0])
        int8[0] = 42
        print(int8[0])  # 42
        assert int8[0] == 42
        # print(int8.length)  # 2
        assert int8.length == 2
        # print(int8.BYTES_PER_ELEMENT)  # 1
        assert int8.BYTES_PER_ELEMENT == 1

        # From an array
        arr = Int8Array([21, 31])
        # print(arr[1])  # 31
        assert arr[1] == 31

        # From another TypedArray
        x = Int8Array([21, 31])
        y = Int8Array(x)
        # print(y[0])  # 21
        assert y[0] == 21

        # From an ArrayBuffer
        b = ArrayBuffer(8)
        z = Int8Array(b, 1, 4)
        # print(z[0])  # 0
        print(z.length)
        assert z[0] == 0
        assert z[1] == 0
        assert z[2] == 0
        assert z[3] == 0
        assert z.length == 4

        # test Int8Array in various ways

        # From a length
        int8 = Int8Array(2)
        # print(int8[0])
        # print(Int8Array(2)[0])
        # print(Int8Array(25)[0])

        # From an array
        arr = Int8Array([21, 31])
        # print(arr[1])  # 31
        assert arr[1] == 31

        # From another TypedArray
        x = Int8Array([21, 31])
        y = Int8Array(x)
        # print(y[0])  # 21
        assert y[0] == 21

        # From an ArrayBuffer
        # b = ArrayBuffer(8)
        # z = Int8Array(b, 1, 4)

        # print(z[0])  # 31
        # assert z[0] == 31

        # From a DataView
        # d = DataView(b, 1, 4)
        # print(d.getInt8(0))  # 31
        # assert d.getInt8(0) == 31

        # var size = 1000000;
        # var buffer = new ArrayBuffer(4 * size);
        # var intArray = new Int32Array(buffer);
        # var array = new Array(size);
        # // Ensure all values are 0 in the array.
        # for (var i = 0; i < size; i++) {
        #     array[i] = 0;
        # }

        size = 1000000
        # size = 10
        buf = ArrayBuffer(4 * size)
        intArray = Int32Array(buf)
        array = [0] * size
        for i in range(size):
            array[i] = 0

        # assert various features of the array
        # print(intArray.length)  # 1000000
        # assert intArray.length == 1000000
        # print(intArray.BYTES_PER_ELEMENT)  # 4
        assert intArray.BYTES_PER_ELEMENT == 4
        # print(intArray.byteLength)  # 4000000
        assert intArray.byteLength == 4000000
        # print(intArray.byteOffset)  # 0
        assert intArray.byteOffset == 0


        size = 100
        intArray = Int8Array(4 * size)
        array = [0] * size

        # var x1 = performance.now();
        x1 = time.time()

        # // Version 1: modify values in Int32Array.
        # for (var i = 0; i < 1000; i++) {
        #     for (var z = 0; z < size; z++) {
        #         intArray[z] = intArray[z] + 2;
        #     }
        # }
        for i in range(100):
            for z in range(size):
                intArray[z] = intArray[z] + 2

        # var x2 = performance.now();
        x2 = time.time()

        # // Version 2: modify values in Array.
        # for (var i = 0; i < 1000; i++) {
        #     for (var z = 0; z < size; z++) {
        #         array[z] = array[z] + 2;
        #     }
        # }
        for i in range(100):
            for z in range(size):
                array[z] = array[z] + 2

        # var x3 = performance.now();
        x3 = time.time()
        # // Results.
        print("TIME 1: " , (x2 - x1))
        print("TIME 2: " , (x3 - x2))

        # test uint8array in various ways
        arr = Uint8Array()
        print(arr)
        assert arr.length == 0

        arr = Uint8Array(2)
        print(arr.length)
        assert arr.length == 2

        # test uint8array in various ways
        arr = Uint8ClampedArray()
        print(arr)
        assert arr.length == 0

        arr = Uint8ClampedArray(2)
        print(arr.length)
        assert arr.length == 2

        # test uint8array in various ways
        arr = Int16Array()
        print(arr)
        assert arr.length == 0

        arr = Int16Array(2)
        print(arr.length)
        assert arr.length == 2

        # test uint8array in various ways
        arr = Uint16Array()
        print(arr)
        assert arr.length == 0

        arr = Uint16Array(2)
        print(arr.length)
        assert arr.length == 2

        # test uint8array in various ways
        arr = Int32Array()
        print(arr)
        assert arr.length == 0

        arr = Int32Array(2)
        print(arr.length)
        assert arr.length == 2

        # test uint8array in various ways
        arr = Uint32Array()
        print(arr)
        assert arr.length == 0

        arr = Uint32Array(2)
        print(arr.length)
        assert arr.length == 2

        # test uint8array in various ways
        arr = Float32Array()
        print(arr)
        assert arr.length == 0

        arr = Float32Array(2)
        print(arr.length)
        assert arr.length == 2

        # test uint8array in various ways
        arr = Float64Array()
        print(arr)
        assert arr.length == 0

        arr = Float64Array(2)
        print(arr.length)
        assert arr.length == 2

    # def test_storage(self):
    #     print("test_storage")

    #     myObj = Storage()
    #     myObj.name = 'John'
    #     myObj.age = 30
    #     myObj.address = '123 Main St'

    #     assert myObj.name == 'John'
    #     assert myObj.age == 30
    #     assert myObj.address == '123 Main St'

    #     window.localStorage.setItem('myObj', myObj)
    #     window.localStorage.getItem('myObj')
    #     myObj2 = window.localStorage.getItem('myObj')
    #     assert myObj2.name == 'John'
    #     assert myObj2.age == 30


    # def test_reflect(self):
    #     print("test_reflect")
    #     myObj = {'name': 'John', 'age': 30, 'address': '123 Main St'}

    #     myObj2 = Reflect(myObj)
    #     assert myObj2.name == 'John'
    #     assert myObj2.age == 30
    #     assert myObj2.address == '123 Main St'
    #     assert myObj2.toString() == '{"name": "John", "age": 30, "address": "123 Main St"}'


    # def test_symbol(self):
    #     print("test_symbol")
    #     symbol = Symbol('x')

    #     assert symbol.name == 'x'
    #     assert symbol.value == 0
    #     assert symbol.toString() == 'x'
    #     assert symbol.toNumber() == 0
    #     assert symbol.toString(True) == '0'


_intID = None
_results = []


if __name__ == '__main__':
    unittest.main()



"""
    domonic.geom.vec3
    ====================================
    written by.ai
"""
import math

# from domonic.javascript import Math


class vec3:
    """[vec3]"""

    def __init__(self, x: float = 0, y: float = 0, z: float = 0, w: float = 0):
        self.x: float = x
        self.y: float = y
        self.z: float = z
        self.w: float = w

    def __add__(self, other):
        if isinstance(other, vec3):
            return self.__class__(
                (self.x + other.x, self.y + other.y, self.z + other.z, self.w + other.w)
            )
        else:
            return self.__class__(
                (self.x + other, self.y + other, self.z + other, self.w + other)
            )

    def __sub__(self, other):
        if isinstance(other, vec3):
            return self.__class__(
                (self.x - other.x, self.y - other.y, self.z - other.z, self.w - other.w)
            )
        else:
            return self.__class__(
                (self.x - other, self.y - other, self.z - other, self.w - other)
            )

    def __mul__(self, other):
        if isinstance(other, vec3):
            return self.__class__(
                (self.x * other.x, self.y * other.y, self.z * other.z, self.w * other.w)
            )
        else:
            return self.__class__(
                (self.x * other, self.y * other, self.z * other, self.w * other)
            )

    # def __rmul__(self, other):
    #     return vec3(other.x * self.x, other.y * self.y, other.z * self.z, other.w * self.w)

    def __truediv__(self, other):
        if isinstance(other, vec3):
            return self.__class__(
                self.x / other.x, self.y / other.y, self.z / other.z, self.w / other.w
            )
        else:
            return self.__class__(
                self.x / other, self.y / other, self.z / other, self.w / other
            )

    # def __pow__(self, other):
    #     return vec3(self.x ** other.x, self.y ** other.y, self.z ** other.z, self.w ** other.w)

    # def __mod__(self, other):
    #     return vec3(self.x % other.x, self.y % other.y, self.z % other.z, self.w % other.w)

    def __getitem__(self, item):
        if isinstance(item, int):
            if item == 0:
                return self.x
            elif item == 1:
                return self.y
            elif item == 2:
                return self.z
            elif item == 3:
                return self.w
        elif isinstance(item, str):
            if item == "x":
                return self.x
            elif item == "y":
                return self.y
            elif item == "z":
                return self.z
            elif item == "w":
                return self.w

    def __iadd__(self, other):
        self.x += other.x
        self.y += other.y
        self.z += other.z
        self.w += other.w

    def add(self, point):
        self.x += point.x
        self.y += point.y
        self.z += point.z
        return self

    def subtract(self, point):
        """Subtract from this point."""
        self.x -= point.x
        self.y -= point.y
        self.z -= point.z
        return self

    def dot(self, other):
        return self.x * other.x + self.y * other.y + self.z * other.z + self.w * other.w

    def cross(self, other):
        return (
            self.x * other.y - self.y * other.x,
            self.y * other.z - self.z * other.y,
            self.z * other.x - self.x * other.z,
        )

    def mul(self, v):
        return v.x * self.x + v.y * self.y + v.z * self.z + v.w * self.w

    def copy(self):
        """Creates a copy of this object."""
        return vec3(self.x, self.y, self.z, self.w)

    def angleBetween(self, other):
        dx = self.x - other.x
        dy = self.y - other.y
        dz = self.z - other.z
        return math.sqrt(dx * dx + dy * dy + dz * dz)

    def length(self):
        return math.sqrt(
            self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w
        )

    def distance(self, other):
        """Returns the distance between this point and another vector3."""
        return (
            (self.x - other.x) ** 2
            + (self.y - other.y) ** 2
            + (self.z - other.z) ** 2
            + (self.w - other.w) ** 2
        )
        # return sqrt(pow(self.x - other.x, 2) + pow(self.y - other.y, 2) + pow(self.z - other.z, 2) + pow(self.w - other.w, 2))

    def equals(self, other):
        """Determine whether two objects are identical."""
        return (
            self.x == other.x
            and self.y == other.y
            and self.z == other.z
            and self.w == other.w
        )

    def intersects(self, other):
        dx = self.x - other.x
        dy = self.y - other.y
        dz = self.z - other.z
        dw = self.w - other.w
        return (
            (dx * dy - dy * dz) > 0
            and (dx * dw - dw * dz) > 0
            and (dx * dz + dz * dw) > 0
            and (dy * dw - dw * dy) > 0
            and (dy * dw - dw * dw) > 0
        )

    def clone(self):
        """Returns a new instance of this vector3."""
        return vec3(self.x, self.y, self.z, self.w)

    def apply(self, point, amount):
        """Moves the points x,y,z,w by amount."""
        return vec3(
            point.x + amount.x,
            point.y + amount.y,
            point.z + amount.z,
            point.w + amount.w,
        )

    def __str__(self):
        return str(self.x) + " " + str(self.y) + " " + str(self.z)



"""
    domonic.i18n.it
    ====================================

    html tags in Italian

    first translated to Italian using google translate.
    send a pull request if you have a better translation.
"""
from domonic.html import *

corpo = body
testa = head
# script = script
stile = style
portale = portal
forte = strong
citazione = blockquote
tabella = table
forma = form
etichetta = label
inviare = submit
titolo = title
noscript = noscript
sezione = section
# nav = nav
articolo = article
aparte = aside
hgruppo = hgroup
indirizzo = address
figura = figure
figcaption = figcaption
piccolo = small
citare = cite
codice = code
sotto = sub
# sup = sup
segno = mark
rubino = ruby
# span = span
# iframe = iframe
video = video
audio = audio
tela = canvas
didascalia = caption
colgroup = colgroup
# tbody = tbody
# testa = thead
tpiede = tfoot
setdicampi = fieldset
legenda = legend
pulsante = button
seleziona = select
listadati = datalist
optgroup = optgroup
opzione = option
areaditesto = textarea
uscita = output
progresso = progress
metro = meter
dettagli = details
riassunto = summary
menù = menu
vocedelmenu = menuitem
carattere = font
intestazione = header
pièdipagina = footer
# base = base
collegamento = link
# meta = meta
# img = img
incorporare = embed
fonte = source
traccia = track
# area = area
# col = col
ingresso = input
comando = command
principale = main
caratteredibase = basefont
centro = center
incorporare = embed
carattere = font
isindice = isindex
lista = listing
testoinchiaro = plaintext
sciopero = strike
modello = template
immagine = picture
# doctype = doctype
commento = comment



"""
    domonic.aframe
    ====================================

    Generate aframe tags with python 3
    https://aframe.io/

"""

from domonic.dom import Element


class aframe_tag(Element):
    def __str__(self):
        return f"<a-{self.name}{self.__attributes__}>{self.content}</a-{self.name}>"


entity = type('entity', (aframe_tag, Element,), {'name': 'entity'})
scene = type('scene', (aframe_tag, Element,), {'name': 'scene'})
material = type('material', (aframe_tag, Element,), {'name': 'material'})
appearance = type('appearance', (aframe_tag, Element,), {'name': 'appearance'})

shape = type('shape', (aframe_tag, Element,), {'name': 'shape'})
# transform = type('transform', (aframe_tag, Element,), {'name': 'transform'})
# inline = type('inline', (aframe_tag, Element,), {'name': 'inline'})

sphere = type('sphere', (aframe_tag, Element,), {'name': 'sphere'})
box = type('box', (aframe_tag, Element,), {'name': 'box'})
plane = type('plane', (aframe_tag, Element,), {'name': 'plane'})
sky = type('sky', (aframe_tag, Element,), {'name': 'sky'})

mixin = type('mixin', (aframe_tag, Element,), {'name': 'mixin'})
circle = type('circle', (aframe_tag, Element,), {'name': 'circle'})
camera = type('camera', (aframe_tag, Element,), {'name': 'camera'})
cone = type('cone', (aframe_tag, Element,), {'name': 'cone'})
cursor = type('cursor', (aframe_tag, Element,), {'name': 'cursor'})

curvedimage = type('curvedimage', (aframe_tag, Element,), {'name': 'curvedimage'})
cylinder = type('cylinder', (aframe_tag, Element,), {'name': 'cylinder'})

dodecahedron = type('dodecahedron', (aframe_tag, Element,), {'name': 'dodecahedron'})
icosahedron = type('icosahedron', (aframe_tag, Element,), {'name': 'icosahedron'})
image = type('image', (aframe_tag, Element,), {'name': 'image'})
light = type('light', (aframe_tag, Element,), {'name': 'light'})
alink = type('alink', (aframe_tag, Element,), {'name': 'alink'})
octahedron = type('octahedron', (aframe_tag, Element,), {'name': 'octahedron'})
plane = type('plane', (aframe_tag, Element,), {'name': 'plane'})
ring = type('ring', (aframe_tag, Element,), {'name': 'ring'})
sound = type('sound', (aframe_tag, Element,), {'name': 'sound'})
sphere = type('sphere', (aframe_tag, Element,), {'name': 'sphere'})
tetrahedron = type('tetrahedron', (aframe_tag, Element,), {'name': 'tetrahedron'})
text = type('text', (aframe_tag, Element,), {'name': 'text'})
torus = type('torus', (aframe_tag, Element,), {'name': 'torus'})
triangle = type('triangle', (aframe_tag, Element,), {'name': 'triangle'})
video = type('video', (aframe_tag, Element,), {'name': 'video'})
videosphere = type('videosphere', (aframe_tag, Element,), {'name': 'videosphere'})

gltf = type('gltf', (aframe_tag, Element,), {'name': 'gltf'})
gltf_model = type('gltf_model', (aframe_tag, Element,), {'name': 'gltf_model'})
torus_knot = type('torus_knot', (aframe_tag, Element,), {'name': 'torus_knot'})
obj_model = type('obj_model', (aframe_tag, Element,), {'name': 'obj_model'})

# <a-box>
# <a-camera>
# <a-circle>
# <a-cone>
# <a-cursor>
# <a-curvedimage>
# <a-cylinder>
# <a-dodecahedron>
# <a-gltf-model>
# <a-icosahedron>
# <a-image>
# <a-light>
# <a-link>
# <a-obj-model>
# <a-octahedron>
# <a-plane>
# <a-ring>
# <a-sky>
# <a-sound>
# <a-sphere>
# <a-tetrahedron>
# <a-text>
# <a-torus-knot>
# <a-torus>
# <a-triangle>
# <a-video>
# <a-videosphere>



"""
    domonic.webapi.history
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/History
"""

# from domonic.events import Event, EventTarget


class History:  # (EventTarget):

    def __init__(self, window=None):
        self.window = window
        self.states = []
        self.index = 0
        self.skip_update = False
        try:
            if self.window is not None:
                self.states.append(self.window.location.href)
                self.index = 1
        except Exception as e:
            print('History is empty:', e)

    def _update(self, url: str):
        """ Updates the current history state """
        if self.skip_update:
            self.skip_update = False
            return
        self.states.append(url)
        self.index = len(self.states) - 1
        # self.dispatchEvent(Event('popstate'))

    def back(self):
        """ Loads the previous URL in the history list """
        self.go(-1)
        self.skip_update = True
        if self.window:
            self.window.location = self.states[self.index]
        # self.dispatchEvent(Event('popstate'))

    def forward(self):
        """ Loads the next URL in the history list """
        self.go(1)
        self.skip_update = True
        if self.window:
            self.window.location = self.state  # s[self.index]

    def go(self, n: int):
        """ Loads a specific URL from the history list """
        self.index += n
        if self.index < 1:
            self.index = 1
        if self.index > self.length:
            self.index = self.length
        return self.index

    @property
    def length(self):
        """ Returns the number of URLs in the history list """
        return len(self.states)

    # def pushState(self, data, title, url):
    #     return self.replaceState(data, title, url)

    # def replaceState(self, data, title, url):
    #     self.states[self.index] = url

    @property
    def state(self):
        return self.states[self.index]

    def __repr__(self) -> str:
        # show the full history
        from pprint import pformat
        return pformat(self.states)



"""
    test_domonic
    ~~~~~~~~~~~~
    domonic core tests for __init__.py at the root of the domonic package

    (previously no tests. was using examples/parsing/page.py to drive dev)

"""

import unittest

# from domonic.dom import *
from domonic import domonic

# import requests
# from mock import patch



class TestCase(unittest.TestCase):
    """ Tests for the domonic """

    def test_load(self):
        t1 = domonic.load('<html></html>')
        print(t1)

    def test_loads(self):
        # t1 = domonic.loads('<html></html>')
        # print(t1)
        pass

    def parse(self):
        t1 = domonic.parse('<html></html>')
        assert t1 == "html(),"  # hmm wondering if parse is correct term. as returns pyml strings

        # t2 = domonic.parse('<html><body></body></html>')
        # print(t2)
        # t3 = domonic.parse('<html><body><p></p></body></html>')
        # print(t3)

        pass

    def evaluate(self):
        t1 = domonic.evaluate('<html></html>')
        print(t1)

    def test_hacked_expat_parser(self):
        # test the  hacked version of the xpat parser
        print("test!>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")
        # t1 = domonic.parseString('<html></html>')
        # print(t1)
        t1 = domonic.parseString('''<html><head><link rel="stylesheet" href="https://unpkg.com/marx-css/css/marx.min.css" /><script src="https://code.jquery.com/jquery-3.5.1.min.js"></script><script>
	function add(){
		$('#results').html( Number($('#a').val()) + Number($('#b').val()) )};
</script></head><body><article><div><label>Add numbers:</label><input id="a" /><span>+</span><input id="b" /><button id="calculate_button" onclick="add();">Calculate</button><div>Result:<div id="results"></div></div></div></article></body></html>''')
        print('RES:', t1)
        print('RES:', type(t1))
        print('RES:', str(t1))
        print(t1.getElementById('a'))

        # print(t1)
        return
#         return

        # print( ':FIRE:', type(t1))
        # return
        # print(str(t1))
        # from domonic import render
        # print( render( t1 ) )

        print("test222!")
        t1 = domonic.parseString('<div></div>')
        print(t1)
        print(str(t1))
        # from domonic import render
        # print( render( t1 ) )


if __name__ == '__main__':
    unittest.main()



"""
    domonic.i18n.hi
    ====================================

    html tags in Hindi

    first translated to Hindi using google translate.
    send a pull request if you have a better translation.
"""
from domonic.html import *

शरीर = shareer = body
सिर = sir = head
लिपि = lipi = script
शैली = shailee = style
पोर्टल = portal = portal
मजबूत = majaboot = strong
ब्लॉककोट = blokakot = blockquote
टेबल = tebal = table
फॉर्म = phorm = form
लेबल = lebal = label
सबमिटकरें = sabamit = submit
शीर्षक = sheershak = title
नोस्क्रिप्ट = noskript = noscript
खंड = khand = section
नव = nav = nav
लेख = lekh = article
एकतरफ = ek = aside
एचग्रुप = echagrup = hgroup
पता = pata = address
आकृति = aakrti = figure
अंजीर = anjeer = figcaption
छोटा = chhota = small
उद्धरण = uddharan = cite
कोड = kod = code
उप = up = sub
समर्थन = samarthan = sup
निशान = nishaan = mark
माणिक = maanik = ruby
अवधि = avadhi = span
आईफ्रेम = aaeephrem = iframe
वीडियो = veediyo = video
ऑडियो = odiyo = audio
कैनवास = kainavaas = canvas
कैप्शन = kaipshan = caption
कोलग्रुप = kolagrup = colgroup
शरीर = shareer = tbody
थैड = thaid = thead
टफ़ुट = tafut = tfoot
फ़ील्डसेट = feeldaset = fieldset
किंवदंती = kinvadantee = legend
बटन = batan = button
चुनें = chunen = select
डेटालिस्ट = detaalist = datalist
ऑप्टग्रुप = optagrup = optgroup
विकल्प = vikalp = option
टेक्सटेरिया = teksateriya = textarea
आउटपुट = aautaput = output
प्रगति = pragati = progress
मीटर = meetar = meter
विवरण = vivaran = details
सारांश = saaraansh = summary
मेनू = menoo = menu
मेन्युइटम = menyuitam = menuitem
फ़ॉन्ट = font = font
हेडर = hedar = header
पादलेख = paad = footer
आधार = aadhaar = base
लिंक = link = link
मेटा = meta = meta
आईएमजी = aaeeemajee = img
एम्बेड = embed = embed
स्रोत = srot = source
ट्रैक = traik = track
क्षेत्र = kshetr = area
कर्नल = karnal = col
इनपुट = inaput = input
आदेश = aadesh = command
मुख्य = mukhy = main
बेसफॉन्ट = besaphont = basefont
केंद्र = kendr = center
एम्बेड = embed = embed
फ़ॉन्ट = font = font
आइसइंडेक्स = aaisindeks = isindex
लिस्टिंग = listing = listing
सादापाठ = saada = plaintext
हड़ताल = hadataal = strike
टेम्पलेट = tempalet = template
चित्र = chitr = picture
सिद्धांत = siddhaant = doctype
टिप्पणी = tippanee = comment



"""
    domonic.webapi.xpath
    ====================================
    https://developer.mozilla.org/en-US/docs/Glossary/XPath

    uses elementpath lib.

    TODO - content strings must be TextNodes for it to work.
        so will have to iterate and update them. i.e. Treewalker

"""

import json
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Union

import elementpath


class XPathEvaluator:

    def __init__(self) -> None:
        pass

    def createExpression(self, expression: str):  # , namespaces: Dict[str, str]) -> None:
        return XPathExpression(expression)


class XPathException:
    def __init__(self) -> None:
        pass


class XPathExpression:

    def __init__(self, expr: str):  #, resolver):

        # TODO - hack.
        # need to allow non underscore accessors to get underscored.
        # when that's fixed can remove this.
        expr = expr.replace('[@', '[@_')
        expr = expr.replace('[@__', '[@_')
        expr = expr.replace('(@', '(@_')
        expr = expr.replace('(@__', '(@_')
        expr = expr.replace('/@', '/@_')
        expr = expr.replace('/@__', '/@_')

        if len(expr) <= 0:
            raise Exception('no expression')
        self.selector = elementpath.Selector(expr)

    # TODO - DRY - make some utils . just stole this from Treewalker.
    @staticmethod
    def _upgrade_dom(node):
        def upgrade(el):
            from domonic.dom import Text
            if isinstance(el, (Text, str)):
                return
            for child in el:
                if isinstance(child, str):
                    newchild = Text(child)
                    el.replaceChild(newchild, child)
                    newchild.parentNode = el
        node._iterate(node, upgrade)
        return node

    def evaluate(self, node, type=6):  # XPathResult.ANY_TYPE):
        # note: otherwise would fail on regular text?
        node = XPathExpression._upgrade_dom(node)
        return XPathResult(self.selector.select(node), type)


class XPathNSResolver:

    def __init__(self) -> None:
        pass


class XPathResult:

    ANY_TYPE = 0
    NUMBER_TYPE = 1
    STRING_TYPE = 2
    BOOLEAN_TYPE = 3
    UNORDERED_NODE_ITERATOR_TYPE = 4
    ORDERED_NODE_ITERATOR_TYPE = 5
    UNORDERED_NODE_SNAPSHOT_TYPE = 6
    ORDERED_NODE_SNAPSHOT_TYPE = 7
    ANY_UNORDERED_NODE_TYPE = 8
    FIRST_ORDERED_NODE_TYPE = 9

    def __init__(self, value, _type):
        if _type == XPathResult.ANY_TYPE:
            tov = type(value)
            if tov == 'object':
                _type = self.UNORDERED_NODE_ITERATOR_TYPE
            if tov == 'boolean':
                _type = self.BOOLEAN_TYPE
            if tov == 'string':
                _type = self.STRING_TYPE
            if tov == 'number':
                _type = self.NUMBER_TYPE

        if _type < self.NUMBER_TYPE or self.FIRST_ORDERED_NODE_TYPE < _type:
            raise Exception(f'unknown type: {_type}')

        self.resultType = _type

        if _type == self.NUMBER_TYPE:
            # self.numberValue=value.number() if getattr(value,'isNodeSet',None) else toNumber(value)
            if getattr(value, 'isNodeSet', None):
                self.numberValue = value  # .number()
            else:
                self.numberValue = float(value)
        elif _type == self.STRING_TYPE:
            # self.stringValue=value.string() if getattr(value,'isNodeSet',None) else toString(value)
            if getattr(value, 'isNodeSet', None):
                self.stringValue = value  # .string()
            else:
                self.stringValue = str(value)
        elif _type == self.BOOLEAN_TYPE:
            # self.booleanValue=value.bool() if getattr(value,'isNodeSet',None) else toBoolean(value)
            if getattr(value, 'isNodeSet', None):
                self.booleanValue = value  # .bool()
            else:
                self.booleanValue = bool(value)
        elif _type == self.ANY_UNORDERED_NODE_TYPE or _type == self.FIRST_ORDERED_NODE_TYPE:
            self.singleNodeValue = value  #.first()
        else:
            self.nodes = value  # .list()
            self.snapshotLength = len(value)
            self.index = 0
            self.invalidIteratorState = False

    # def iterateNext(self):
    #     node = self.nodes[self.index]
    #     self.index += 1
    #     return node

    # def snapshotItem(self, i):
    #     return self.nodes[i]



"""
    domonic.window
    ====================================

    be mindful there are 2 types of window to be found in domonic:

        1. the javascript window - a window with only static js methods

        2. the domonic window (this one) - a window connected to other things i.e. dom

    You can extend or import either for your own purposes.

"""
import sys

import requests

from domonic import domonic
from domonic.dom import *
from domonic.dom import Location, document
from domonic.javascript import Window
from domonic.webapi.console import Console
from domonic.webapi.credentials import CredentialsContainer
from domonic.webapi.geo import Geolocation
from domonic.webapi.netinfo import NetworkInformation
from domonic.webapi.webstorage import Storage
from domonic.webapi.history import History

# from domonic.webapi.mediacapabilities import MediaCapabilities
# from domonic.webapi.mediasession import MediaSession


# TODO - test
class CustomElementRegistry():
    """ The CustomElementRegistry interface provides methods for registering custom elements and querying registered elements.
    To get an instance of it, use the window.customElements property. """

    def __init__(self) -> None:
        self.store = {}

    # Defines a new custom element.
    def define(self, name: str, constructor, options=None) -> None:
        """[defines a new custom element.]

        Args:
            name ([str]): [Name for the new custom element. Note that custom element names must contain a hyphen.]
            constructor ([type]): [Constructor for the new custom element.]
            options ([dict]): [Object that controls how the element is defined. One option is currently supported: extends]
        """
        if '-' not in name:
            raise ValueError('Invalid custom element name. Must contain hypen: ' + name)
        # el = document.createElement(name)
        # el.constructor = constructor
        from domonic.dom import Element
        from domonic.html import tag
        el = type(name, (tag, Element), {'name': name, '__init__': constructor})
        if options is not None:
            if 'extends' in options:
                el.extends = options['extends']
        self.store[name] = el
        return el

    def get(self, name):
        """
            Returns the constructor for the named custom element,
            or undefined if the custom element is not defined.
        """
        # see if its in the store or return none
        if name in self.store:
            return self.store[name]
        else:
            return None

    # Upgrades a custom element directly, even before it is connected to its shadow root.
    def upgrade(self):
        pass

    # Returns an empty promise that resolves when a custom element becomes defined with the given name.
    # If such a custom element is already defined, the returned promise is immediately fulfilled.
    def whenDefined(self):
        pass


class Navigator:
    """ Navigator """

    # Determines whether cookies are enabled in the browser
    cookieEnabled = False

    # Determines whether the browser is online
    # onLine = False
    @property
    def onLine(self):
        raise NotImplementedError

    # Returns the name of the browser Navigator
    appName = "domonic"

    def __init__(self, *args, **kwargs):
        self.connection: NetworkInformation = NetworkInformation()
        self.credentials: CredentialsContainer = CredentialsContainer()
        self.geolocation: Geolocation = Geolocation()
        self.hid = None
        self.keyboard = None
        self.locks = None
        self.mediaCapabilities = None
        self.mediaSession = None
        self.mediaDevices = None
        self.presentation = None
        self.serial = None
        self.serviceWorker = None
        self.storage = None
        self.vendor = None
        self.webdriver = None
        self.xr = None
        self.buildID = None
        self.contacts = None
        self._screen = Screen()

    # @property
    # def appVersion():
        """ Returns the version information of the browser """
        # from domonic import __version__
        # return __version__

    # @property
    # def language():
        """ Returns the language of the browser Navigator """
        # import locale
        # return locale.getdefaultlocale()

    # def languages

    @property
    def platform(self) -> str:
        """ Returns the platform """
        if 'darwin' in sys.platform:
            return 'mac'
        elif 'linux' in sys.platform:
            return 'linux'
        elif 'win32' in sys.platform:
            return 'windows'
        else:
            return 'unknown'

    @property
    def product(self) -> str:
        """ Returns the product name """
        return self.appName

    @property
    def userAgent(self) -> str:
        """ Returns the user-agent header sent by the browser to the server Navigator """
        raise NotImplementedError

    @property
    def deviceMemory(self) -> float:
        """ Returns the amount of memory available on the device """
        return 1

    @property
    def doNotTrack(self):
        """ Returns the value of the doNotTrack attribute of the Navigator object """
        # return False
        return 'lol'

    @property
    def hardwareConcurrency(self):
        """ Returns the number of logical processors available to the browser Navigator """
        return 1

    @property
    def maxTouchPoints(self):
        """ Returns the maximum number of touch points Navigator """
        return 1

    @staticmethod
    def registerProtocolHandler(scheme, url, title):
        """ Registers a new protocol handler Navigator """
        raise NotImplementedError

    @staticmethod
    def requestMediaKeySystemAccess(keySystem, supportedConfigurations):
        """ Requests a new MediaKeySystemAccess object Navigator """
        raise NotImplementedError

    def canShare(self):
        """ Returns whether the browser Navigator can share files """
        return False

    def clearAppBadge(self):
        """ Clears the app badge Navigator """
        raise NotImplementedError

    def getBattery(self):
        """ Returns the battery information Navigator """
        raise NotImplementedError

    @property
    def javaEnabled(self):
        """ Returns whether the browser Navigator supports Java """
        return False

    def vibrate(self, pattern):
        """ Vibrates the device Navigator """
        raise NotImplementedError

    # deprecated
    # Navigator.securitypolicy
    # Navigator.standalone
    # Navigator.wakeLock
    # Navigator.appCodeName
    # Navigator.appName
    # Navigator.appVersion
    # Navigator.activeVRDisplays
    # Navigator.battery
    # Navigator.mimeTypes
    # Navigator.oscpu
    # Navigator.platform
    # Navigator.plugins
    # Navigator.product
    # Navigator.productSub
    # Navigator.vendorSub


class Screen(EventTarget):
    # https://developer.mozilla.org/en-US/docs/Web/API/Screen

    def __init__(self):
        """ Screen. (you will need to set them) """
        self.availLeft = 0
        self.availTop = 0
        self.availHeight = 0
        self.availWidth = 0
        self.colorDepth = 24
        self.height = 768
        self.left = 0
        self.pixelDepth = 24
        self.top = 0
        self.width = 1024
        self.orientation = None

    # def lockOrientation(self):
    #     """ Locks the screen orientation """
    #     raise NotImplementedError

    # def unlockOrientation(self):
    #     """ Unlocks the screen orientation """
    #     raise NotImplementedError


class Window(Window):

    def __init__(self):
        self.customElements = CustomElementRegistry()
        self._localStorage: Storage = Storage()  # TODO - should persist across sessions
        self._sessionStorage: Storage = Storage()  # TODO - should reset on page refresh
        self._navigator: Navigator = Navigator()
        self._location: Location = Location('eventual.technology')
        self._console: Console = Console()
        self._history: History = History(self)
        # personalbar?
        super(Window, Window).__init__(self)

    @property
    def history(self) -> History:
        """ Returns the history object """
        return self._history

    @property
    def console(self) -> Console:
        """ Returns the console object """
        return self._console

    @property
    def localStorage(self) -> Storage:
        """ Returns the local storage object """
        return self._localStorage

    @property
    def sessionStorage(self) -> Storage:
        """ Returns the session storage object """
        return self._sessionStorage

    @staticmethod
    def document(self) -> Document:
        from domonic.dom import document
        return document

    @property
    def location(self) -> Location:
        return self._location
        # return Location(self.window.location.href)

    def _set_location_using_htm5lib(self, url):
        # from html5lib import parse
        # return parse(html)
        # import html5lib
        # import requests
        from html5lib import HTMLParser

        from domonic.ext.html5lib_ import getTreeBuilder

        if 'http' not in url:
            url = 'https://' + url

        r = requests.get(url)
        parser = HTMLParser(tree=getTreeBuilder())
        page = parser.parse(r.text)

        # from domonic.dom import document
        document = page
        return document

    @location.setter
    def location(self, value):
        # NOTE - not documented. still unverified
        # self._location = value
        # TODO - load the content of the location using requests

        # self._history.pushState(None, None, value)
        self._history._update(value)  #pushState(None, None, value)

        try:
            import html5lib
            if 'html5lib' in sys.modules:
                self.document = self._set_location_using_htm5lib(value)
                self._location = Location(value)
                self.document.URL = value
                return
        except ImportError:
            pass

        if value is None:
            return
        import requests
        r = requests.get(value)
        content = r.text  # .encode('utf-8')
        # print(content)
        # content = content.replace('\n', '')
        # content = content.replace('\t', '')
        # content = content.replace(' ', '')
        # content = content.replace('\r', '')

        from domonic.parsers import remove_doctype, remove_tags
        content = remove_tags(content, ['js', 'css', '#'])
        content = remove_doctype(content)

        # from domonic.utils import Utils
        # content = Utils.escape(content)
        # content = domonic.parsers.remove_html_tag_by_name(content, 'head') # TODO - buggy.removed start of docs too?
        # content = domonic.parsers.remove_html_tag_by_name(content, 'body')

        # clean invalid html
        content.replace('&', '&amp;')
        content = domonic.parsers.remove_whitespace(content)
        content = domonic.parsers.remove_newlines(content)
        content = domonic.parsers.remove_tabs(content)

        print(content)
        # return
        # clean the html before parsing

        # TODO - don't use parseString as it is not a HTML parser its XML parser. atm I'm just using for testing
        self.document = domonic.domonic.parseString(content)
        self.document.URL = value
        self._location = Location(value)

    def blur(self):
        """ Removes focus from an element """
        raise NotImplementedError

    def closed(self):
        """[Returns a Boolean value indicating whether a window has been closed or not]
        """
        raise NotImplementedError

    def close(self):
        """[Closes the output stream previously opened with document.open()]
        """
        raise NotImplementedError

    def confirm(self, message: str):
        """[Displays a dialog box with a message and an OK and a Cancel button.]
        (https://developer.mozilla.org/en-US/docs/Web/API/Window/confirm)

        Args:
            message ([type]): [the message to display in the dialog box]
        """
        raise NotImplementedError

    @property
    def defaultStatus(self):
        """ Returns the default status message of the window """
        raise NotImplementedError

    @defaultStatus.setter
    def defaultStatus(self, value=None):
        """ Sets the default text in the statusbar of a window """
        raise NotImplementedError

    def focus(self):
        """ Gives focus to an element """
        raise NotImplementedError

    def frameElement(self):
        """Returns the <iframe> element in which the current window is inserted"""
        raise NotImplementedError

    def getComputedStyle(self, el, pseudo=None):
        """ Gets the current computed CSS styles applied to an element """
        raise NotImplementedError

    def getSelection(self):
        """ Returns a Selection object representing the range of text selected by the user """
        raise NotImplementedError

    # def history(self):
        # """ Returns the History object for the window """
        # raise NotImplementedError
        # from domonic.webapi.history import History
        # return History()
        # return self._history

    def innerHeight(self):
        """[Returns the height of the window's content area (viewport) including scrollbars]
        """
        raise NotImplementedError

    def innerWidth(self):
        """[Returns the width of a window's content area (viewport) including scrollbars]
        """
        raise NotImplementedError

    def matchMedia(self, media_query_list):
        """ Returns a MediaQueryList object representing the specified CSS media query string """
        raise NotImplementedError

    def moveBy(self, x: int, y: int):
        """[Moves a window relative to its current position]

        Args:
            x ([int]): [the horizontal offset]
            y ([int]): [the vertical offset]
        """
        raise NotImplementedError

    def moveTo(self, x: int, y: int):
        """[Moves a window to the specified position]

        Args:
            x (int): [the position on the x-axis]
            y (int): [the position on the y-axis]
        """
        raise NotImplementedError

    def name(self):
        """ Returns the name of the window """
        raise NotImplementedError

    @property
    def navigator(self):
        """ Returns the Navigator object for the window """
        return self._navigator

    @property
    def screen(self) -> Screen:
        """Returns the Screen object for the window (See Screen object)"""
        return self._screen

    @property
    def screenLeft(self) -> int:
        """Returns the horizontal coordinate of the window relative to the screen"""
        return self._screen.screenLeft

    @property
    def screenTop(self) -> int:
        """Returns the vertical coordinate of the window relative to the screen"""
        return self._screen.screenTop

    '''
    # # The event occurs when the window's history changes  PopStateEvent?
    # def onpopstate(self):
    #     raise NotImplementedError

    # Opens an HTML output stream to collect output from document.write() Document, Window
    def open(self):
        raise NotImplementedError

    # Returns a reference to the window that created the window
    def opener(self):
        raise NotImplementedError

    # Returns the height of the browser window, including toolbars/scrollbars Window
    def outerHeight(self):
        raise NotImplementedError

    # Returns the width of the browser window, including toolbars/scrollbars
    def outerWidth(self):
        raise NotImplementedError

    # Returns the pixels the current document has been scrolled (horizontally) from the upper left corner of the window
    def pageXOffset(self):
        raise NotImplementedError

    # Returns the pixels the current document has been scrolled (vertically) from the upper left corner of the window Window
    def pageYOffset(self):
        raise NotImplementedError

    # Returns the parent window of the current window Window
    def parent(self):
        raise NotImplementedError

    # Prints the content of the current window
    def _print(self):
        raise NotImplementedError

    # Resizes the window by the specified pixels
    def resizeBy(self):
        raise NotImplementedError

    # Resizes the window to the specified width and height
    def resizeTo(self):
        raise NotImplementedError

    # Deprecated. This method has been replaced by the scrollTo() method. Window
    def scroll(self):
        raise NotImplementedError

    # Scrolls the document by the specified number of pixels
    def scrollBy(self):
        raise NotImplementedError

    # Scrolls the specified element into the visible area of the browser window   Element
    def scrollIntoView(self):
        raise NotImplementedError

    # Scrolls the document to the specified coordinates  < TODO - this will be fun
    def scrollTo(self):
        raise NotImplementedError

    # An alias of pageXOffset Window
    def scrollX(self):
        raise NotImplementedError

    # An alias of pageYOffset Window
    def scrollY(self):
        raise NotImplementedError

    # Stops the window from loading
    def stop(self):
        raise NotImplementedError

    # Sets or returns the text in the statusbar of a window
    def status(self):
        raise NotImplementedError

    # the topmost browser window
    def top(self):
        raise NotImplementedError
    '''


# global window
window = Window()



"""
    test_window
    ~~~~~~~~~~~~~~~~
"""

import unittest

from domonic.decorators import *
# from domonic.dom import *
# from domonic.javascript import *
# from domonic.webapi import *
from domonic.window import *


class TestCase(unittest.TestCase):

    @silence
    def test_window(self):
        window.location = "http://www.google.com"
        # print(window.document.body.innerHTML)
        print(window.document.title)


if __name__ == '__main__':
    unittest.main()



"""
    domonic.mathml
    ====================================

    Generate MATHML using python 3

"""

from domonic.dom import Element

mathml_tags = [
    "mmultiscripts",
    "munderover",
    "semantics",
    "maligngroup",
    "malignmark",
    "mprescripts",
    "munderover",
    "mscarries",
    "maction",
    "menclose",
    "mphantom",
    "mpadded",
    "merror",
    "mfenced",
    "msgroup",
    "mlongdiv",
    "mfrac",
    "mover",
    "mroot",
    "mspace",
    "msqrt",
    "mstyle",
    "msubsup",
    "mtable",
    "mtext",
    "munder",
    "msline",
    "mstyle",
    "mscarry",
    "munder",
    "math",
    "mrow",
    "msub",
    "msup",
    "none",
    "mtr",
    "mtd",
    "ms",
    "mi",
    "mn",
    "mo",
]

# mathml_attributes = []

math_ = type("math", (Element,), {"name": "math"})
maction = type("maction", (Element,), {"name": "maction"})
menclose = type("menclose", (Element,), {"name": "menclose"})
merror = type("merror", (Element,), {"name": "merror"})
mfenced = type("mfenced", (Element,), {"name": "mfenced"})
mfrac = type("mfrac", (Element,), {"name": "mfrac"})
mi = type("mi", (Element,), {"name": "mi"})
mmultiscripts = type("mmultiscripts", (Element,), {"name": "mmultiscripts"})
mn = type("mn", (Element,), {"name": "mn"})
mo = type("mo", (Element,), {"name": "mo"})
mover = type("mover", (Element,), {"name": "mover"})
mpadded = type("mpadded", (Element,), {"name": "mpadded"})
mphantom = type("mphantom", (Element,), {"name": "mphantom"})
mroot = type("mroot", (Element,), {"name": "mroot"})
mrow = type("mrow", (Element,), {"name": "mrow"})
ms = type("ms", (Element,), {"name": "ms"})
mspace = type("mspace", (Element,), {"name": "mspace"})
msqrt = type("msqrt", (Element,), {"name": "msqrt"})
mstyle = type("mstyle", (Element,), {"name": "mstyle"})
msub = type("msub", (Element,), {"name": "msub"})
msubsup = type("msubsup", (Element,), {"name": "msubsup"})
msup = type("msup", (Element,), {"name": "msup"})
mtable = type("mtable", (Element,), {"name": "mtable"})
mtd = type("mtd", (Element,), {"name": "mtd"})
mtext = type("mtext", (Element,), {"name": "mtext"})
mtr = type("mtr", (Element,), {"name": "mtr"})
munder = type("munder", (Element,), {"name": "munder"})
munderover = type("munderover", (Element,), {"name": "munderover"})
semantics = type("semantics", (Element,), {"name": "semantics"})
maligngroup = type("maligngroup", (Element,), {"name": "maligngroup"})
malignmark = type("malignmark", (Element,), {"name": "malignmark"})
msline = type("msline", (Element,), {"name": "msline"})
msgroup = type("msgroup", (Element,), {"name": "msgroup"})
mlongdiv = type("mlongdiv", (Element,), {"name": "mlongdiv"})
mstyle = type("mstyle", (Element,), {"name": "mstyle"})
mprescripts = type("mprescripts", (Element,), {"name": "mprescripts"})
mscarries = type("mscarries", (Element,), {"name": "mscarries"})
mscarry = type("mscarry", (Element,), {"name": "mscarry"})
munder = type("munder", (Element,), {"name": "munder"})
munderover = type("munderover", (Element,), {"name": "munderover"})
none = type("none", (Element,), {"name": "none"})



"""
    domonic.collada
    ====================================
    Generate Collada with python 3

"""

from domonic.dom import Element
from domonic.html import closed_tag


def write(inp, outp='output'):
    """write

    Renders the input to string or .dae file

    Args:
        inp (obj): domonic.collada tags
        outp (str): An optional filename for the .dae file that gets produced

    Returns:
        str: A HTML rendered string
    """
    if outp != '':
        with open(outp, "w+") as f:
            f.write(str(inp) + ".dae")
    return str(inp)


COLLADA = type('COLLADA', (Element,), {'name': 'COLLADA'})

asset = type('asset', (Element,), {'name': 'asset'})
library_cameras = type('library_cameras', (Element,), {'name': 'library_cameras'})
library_lights = type('library_lights', (Element,), {'name': 'library_lights'})
library_materials = type('library_materials', (Element,), {'name': 'library_materials'})
library_effects = type('library_effects', (Element,), {'name': 'library_effects'})
library_geometries = type('library_geometries', (Element,), {'name': 'library_geometries'})
library_visual_scenes = type('library_visual_scenes', (Element,), {'name': 'library_visual_scenes'})
scene = type('scene', (Element,), {'name': 'scene'})
extra = type('extra', (Element,), {'name': 'extra'})
author = type('author', (Element,), {'name': 'author'})
up_axis = type('up_axis', (Element,), {'name': 'up_axis'})
camera = type('camera', (Element,), {'name': 'camera'})
optics = type('optics', (Element,), {'name': 'optics'})
technique = type('technique', (Element,), {'name': 'technique'})
technique_common = type('technique_common', (Element,), {'name': 'technique_common'})
aspect_ratio = type('aspect_ratio', (Element,), {'name': 'aspect_ratio'})
light = type('light', (Element,), {'name': 'light'})
constant_attenuation = type('constant_attenuation', (Element,), {'name': 'constant_attenuation'})
linear_attenuation = type('linear_attenuation', (Element,), {'name': 'linear_attenuation'})
quadratic_attenuation = type('quadratic_attenuation', (Element,), {'name': 'quadratic_attenuation'})
material = type('material', (Element,), {'name': 'material'})
effect = type('effect', (Element,), {'name': 'effect'})
profile_COMMON = type('profile_COMMON', (Element,), {'name': 'profile_COMMON'})
phong = type('phong', (Element,), {'name': 'phong'})
emission = type('emission', (Element,), {'name': 'emission'})
geometry = type('geometry', (Element,), {'name': 'geometry'})
mesh = type('mesh', (Element,), {'name': 'mesh'})
source = type('source', (Element,), {'name': 'source'})
vertices = type('vertices', (Element,), {'name': 'vertices'})
polylist = type('polylist', (Element,), {'name': 'polylist'})
vcount = type('vcount', (Element,), {'name': 'vcount'})
p = type('p', (Element,), {'name': 'p'})
library_visual_scenes = type('library_visual_scenes', (Element,), {'name': 'library_visual_scenes'})
visual_scene = type('visual_scene', (Element,), {'name': 'visual_scene'})
node = type('node', (Element,), {'name': 'node'})
instance_visual_scene = type('instance_visual_scene', (Element,), {'name': 'instance_visual_scene'})
test_element = type('test_element', (Element,), {'name': 'test_element'})
instance_effect = type('instance_effect', (closed_tag,), {'name': 'instance_effect'})
texture_unit = type('texture_unit', (closed_tag,), {'name': 'texture_unit'})
tapered_cylinder = type('tapered_cylinder', (closed_tag,), {'name': 'tapered_cylinder'})
tapered_capsule = type('tapered_capsule', (closed_tag,), {'name': 'tapered_capsule'})
format_hint = type('format_hint', (closed_tag,), {'name': 'format_hint'})
revisions = type('revisions', (closed_tag,), {'name': 'revisions'})
contributer = type('contributer', (closed_tag,), {'name': 'contributer'})
acceleration = type('acceleration', (Element,), {'name': 'acceleration'})
axis_info = type('axis_info', (Element,), {'name': 'axis_info'})
accessor = type('accessor', (Element,), {'name': 'accessor'})
active = type('active', (Element,), {'name': 'active'})
alpha_func = type('alpha_func', (Element,), {'name': 'alpha_func'})
alpha_test_enable = type('alpha_test_enable', (Element,), {'name': 'alpha_test_enable'})
alpha = type('alpha', (Element,), {'name': 'alpha'})
altitude = type('altitude', (Element,), {'name': 'altitude'})
ambient = type('ambient', (Element,), {'name': 'ambient'})
angle = type('angle', (Element,), {'name': 'angle'})
angular_velocity = type('angular_velocity', (Element,), {'name': 'angular_velocity'})
angular = type('angular', (Element,), {'name': 'angular'})
animation_clip = type('animation_clip', (Element,), {'name': 'animation_clip'})
animation = type('animation', (Element,), {'name': 'animation'})
annotate = type('annotate', (Element,), {'name': 'annotate'})
argument = type('argument', (Element,), {'name': 'argument'})
array = type('array', (Element,), {'name': 'array'})
articulated_system = type('articulated_system', (Element,), {'name': 'articulated_system'})
perspective = type('perspective', (Element,), {'name': 'perspective'})
asset = type('asset', (Element,), {'name': 'asset'})
se = type('se', (Element,), {'name': 'se'})
attachment_end = type('attachment_end', (Element,), {'name': 'attachment_end'})
attachment_full = type('attachment_full', (Element,), {'name': 'attachment_full'})
attachment_start = type('attachment_start', (Element,), {'name': 'attachment_start'})
attachment = type('attachment', (Element,), {'name': 'attachment'})
author_email = type('author_email', (Element,), {'name': 'author_email'})
author_website = type('author_website', (Element,), {'name': 'author_website'})
author = type('author', (Element,), {'name': 'author'})
authoring_tool = type('authoring_tool', (Element,), {'name': 'authoring_tool'})
auto_normal_enable = type('auto_normal_enable', (Element,), {'name': 'auto_normal_enable'})
axis_info = type('axis_info', (Element,), {'name': 'axis_info'})
axis = type('axis', (Element,), {'name': 'axis'})
prismatic = type('prismatic', (Element,), {'name': 'prismatic'})
revolute = type('revolute', (Element,), {'name': 'revolute'})
swept_surface = type('swept_surface', (Element,), {'name': 'swept_surface'})
binary = type('binary', (Element,), {'name': 'binary'})
bind_attribute = type('bind_attribute', (Element,), {'name': 'bind_attribute'})
bind_joint_axis = type('bind_joint_axis', (Element,), {'name': 'bind_joint_axis'})
bind_kinematics_model = type('bind_kinematics_model', (Element,), {'name': 'bind_kinematics_model'})
bind_shape_matrix = type('bind_shape_matrix', (Element,), {'name': 'bind_shape_matrix'})
bind_uniform = type('bind_uniform', (Element,), {'name': 'bind_uniform'})
bind_vertex_input = type('bind_vertex_input', (Element,), {'name': 'bind_vertex_input'})
bind = type('bind', (Element,), {'name': 'bind'})
bind = type('bind', (Element,), {'name': 'bind'})
blend_color = type('blend_color', (Element,), {'name': 'blend_color'})
blend_enable = type('blend_enable', (Element,), {'name': 'blend_enable'})
blend_equation_separate = type('blend_equation_separate', (Element,), {'name': 'blend_equation_separate'})
blend_equation = type('blend_equation', (Element,), {'name': 'blend_equation'})
blend_func_separate = type('blend_func_separate', (Element,), {'name': 'blend_func_separate'})
blend_func = type('blend_func', (Element,), {'name': 'blend_func'})
blinn = type('blinn', (Element,), {'name': 'blinn'})
bool_array = type('bool_array', (Element,), {'name': 'bool_array'})
border_color = type('border_color', (Element,), {'name': 'border_color'})
box = type('box', (Element,), {'name': 'box'})
brep = type('brep', (Element,), {'name': 'brep'})
camera = type('camera', (Element,), {'name': 'camera'})
capsule = type('capsule', (Element,), {'name': 'capsule'})
channel = type('channel', (Element,), {'name': 'channel'})
arget = type('arget', (Element,), {'name': 'arget'})
circle = type('circle', (Element,), {'name': 'circle'})
clip_plane_enable = type('clip_plane_enable', (Element,), {'name': 'clip_plane_enable'})
clip_plane = type('clip_plane', (Element,), {'name': 'clip_plane'})
code = type('code', (Element,), {'name': 'code'})
color_clear = type('color_clear', (Element,), {'name': 'color_clear'})
color_logic_op_enable = type('color_logic_op_enable', (Element,), {'name': 'color_logic_op_enable'})
color_mask = type('color_mask', (Element,), {'name': 'color_mask'})
color_material_enable = type('color_material_enable', (Element,), {'name': 'color_material_enable'})
color_material = type('color_material', (Element,), {'name': 'color_material'})
color_target = type('color_target', (Element,), {'name': 'color_target'})
color = type('color', (Element,), {'name': 'color'})
comments = type('comments', (Element,), {'name': 'comments'})
compiler = type('compiler', (Element,), {'name': 'compiler'})
cone = type('cone', (Element,), {'name': 'cone'})
connect_param = type('connect_param', (Element,), {'name': 'connect_param'})
point = type('point', (Element,), {'name': 'point'})
spot = type('spot', (Element,), {'name': 'spot'})
constant = type('constant', (Element,), {'name': 'constant'})
constant = type('constant', (Element,), {'name': 'constant'})
contributor = type('contributor', (Element,), {'name': 'contributor'})
control_vertices = type('control_vertices', (Element,), {'name': 'control_vertices'})
controller = type('controller', (Element,), {'name': 'controller'})
convex_mesh = type('convex_mesh', (Element,), {'name': 'convex_mesh'})
copyright = type('copyright', (Element,), {'name': 'copyright'})
coverage = type('coverage', (Element,), {'name': 'coverage'})
create_2d = type('create_2d', (Element,), {'name': 'create_2d'})
create_3d = type('create_3d', (Element,), {'name': 'create_3d'})
create_cube = type('create_cube', (Element,), {'name': 'create_cube'})
created = type('created', (Element,), {'name': 'created'})
cull_face_enable = type('cull_face_enable', (Element,), {'name': 'cull_face_enable'})
cull_face = type('cull_face', (Element,), {'name': 'cull_face'})
curve = type('curve', (Element,), {'name': 'curve'})
curves = type('curves', (Element,), {'name': 'curves'})
cylinder = type('cylinder', (Element,), {'name': 'cylinder'})
cylinder = type('cylinder', (Element,), {'name': 'cylinder'})
damping = type('damping', (Element,), {'name': 'damping'})
deceleration = type('deceleration', (Element,), {'name': 'deceleration'})
axis_info = type('axis_info', (Element,), {'name': 'axis_info'})
effector_info = type('effector_info', (Element,), {'name': 'effector_info'})
density = type('density', (Element,), {'name': 'density'})
depth_bounds_enable = type('depth_bounds_enable', (Element,), {'name': 'depth_bounds_enable'})
depth_bounds = type('depth_bounds', (Element,), {'name': 'depth_bounds'})
depth_clamp_enable = type('depth_clamp_enable', (Element,), {'name': 'depth_clamp_enable'})
depth_clear = type('depth_clear', (Element,), {'name': 'depth_clear'})
depth_func = type('depth_func', (Element,), {'name': 'depth_func'})
depth_mask = type('depth_mask', (Element,), {'name': 'depth_mask'})
depth_range = type('depth_range', (Element,), {'name': 'depth_range'})
depth_target = type('depth_target', (Element,), {'name': 'depth_target'})
depth_test_enable = type('depth_test_enable', (Element,), {'name': 'depth_test_enable'})
diffuse = type('diffuse', (Element,), {'name': 'diffuse'})
direction = type('direction', (Element,), {'name': 'direction'})
line = type('line', (Element,), {'name': 'line'})
swept_surface = type('swept_surface', (Element,), {'name': 'swept_surface'})
directional = type('directional', (Element,), {'name': 'directional'})
dither_enable = type('dither_enable', (Element,), {'name': 'dither_enable'})
draw = type('draw', (Element,), {'name': 'draw'})
dynamic_friction = type('dynamic_friction', (Element,), {'name': 'dynamic_friction'})
dynamic = type('dynamic', (Element,), {'name': 'dynamic'})
edges = type('edges', (Element,), {'name': 'edges'})
effect = type('effect', (Element,), {'name': 'effect'})
effector_info = type('effector_info', (Element,), {'name': 'effector_info'})
ellipse = type('ellipse', (Element,), {'name': 'ellipse'})
emission = type('emission', (Element,), {'name': 'emission'})
enabled = type('enabled', (Element,), {'name': 'enabled'})
equation = type('equation', (Element,), {'name': 'equation'})
evaluate_scene = type('evaluate_scene', (Element,), {'name': 'evaluate_scene'})
evaluate = type('evaluate', (Element,), {'name': 'evaluate'})
exact = type('exact', (Element,), {'name': 'exact'})
extra = type('extra', (Element,), {'name': 'extra'})
faces = type('faces', (Element,), {'name': 'faces'})
falloff_angle = type('falloff_angle', (Element,), {'name': 'falloff_angle'})
falloff_exponent = type('falloff_exponent', (Element,), {'name': 'falloff_exponent'})
float_array = type('float_array', (Element,), {'name': 'float_array'})
focal = type('focal', (Element,), {'name': 'focal'})
fog_color = type('fog_color', (Element,), {'name': 'fog_color'})
fog_coord_src = type('fog_coord_src', (Element,), {'name': 'fog_coord_src'})
fog_density = type('fog_density', (Element,), {'name': 'fog_density'})
fog_enable = type('fog_enable', (Element,), {'name': 'fog_enable'})
fog_end = type('fog_end', (Element,), {'name': 'fog_end'})
fog_mode = type('fog_mode', (Element,), {'name': 'fog_mode'})
fog_state = type('fog_state', (Element,), {'name': 'fog_state'})
force_field = type('force_field', (Element,), {'name': 'force_field'})
formula = type('formula', (Element,), {'name': 'formula'})
frame_object = type('frame_object', (Element,), {'name': 'frame_object'})
frame_origin = type('frame_origin', (Element,), {'name': 'frame_origin'})
frame_tcp = type('frame_tcp', (Element,), {'name': 'frame_tcp'})
frame_tip = type('frame_tip', (Element,), {'name': 'frame_tip'})
front_face = type('front_face', (Element,), {'name': 'front_face'})
func = type('func', (Element,), {'name': 'func'})
geographic_location = type('geographic_location', (Element,), {'name': 'geographic_location'})
geometry = type('geometry', (Element,), {'name': 'geometry'})
gravity = type('gravity', (Element,), {'name': 'gravity'})
h = type('h', (Element,), {'name': 'h'})
half_extents = type('half_extents', (Element,), {'name': 'half_extents'})
height = type('height', (Element,), {'name': 'height'})
capsule = type('capsule', (Element,), {'name': 'capsule'})
binary = type('binary', (Element,), {'name': 'binary'})
init_from = type('init_from', (Element,), {'name': 'init_from'})
hint = type('hint', (Element,), {'name': 'hint'})
hollow = type('hollow', (Element,), {'name': 'hollow'})
IDREF_array = type('IDREF_array', (Element,), {'name': 'IDREF_array'})
image = type('image', (Element,), {'name': 'image'})
imager = type('imager', (Element,), {'name': 'imager'})
include = type('include', (Element,), {'name': 'include'})
index_of_refraction = type('index_of_refraction', (Element,), {'name': 'index_of_refraction'})
index = type('index', (Element,), {'name': 'index'})
inertia = type('inertia', (Element,), {'name': 'inertia'})
instance_rigid_body = type('instance_rigid_body', (Element,), {'name': 'instance_rigid_body'})
init_from = type('init_from', (Element,), {'name': 'init_from'})
inline = type('inline', (Element,), {'name': 'inline'})
emantic = type('emantic', (Element,), {'name': 'emantic'})
emantics = type('emantics', (Element,), {'name': 'emantics'})
instance_animation = type('instance_animation', (Element,), {'name': 'instance_animation'})
instance_articulated_system = type('instance_articulated_system', (Element,), {'name': 'instance_articulated_system'})
instance_camera = type('instance_camera', (Element,), {'name': 'instance_camera'})
instance_controller = type('instance_controller', (Element,), {'name': 'instance_controller'})
instance_effect = type('instance_effect', (Element,), {'name': 'instance_effect'})
instance_force_field = type('instance_force_field', (Element,), {'name': 'instance_force_field'})
instance_formula = type('instance_formula', (Element,), {'name': 'instance_formula'})
instance_geometry = type('instance_geometry', (Element,), {'name': 'instance_geometry'})
instance_image = type('instance_image', (Element,), {'name': 'instance_image'})
instance_joint = type('instance_joint', (Element,), {'name': 'instance_joint'})
instance_kinematics_model = type('instance_kinematics_model', (Element,), {'name': 'instance_kinematics_model'})
instance_kinematics_scene = type('instance_kinematics_scene', (Element,), {'name': 'instance_kinematics_scene'})
instance_light = type('instance_light', (Element,), {'name': 'instance_light'})
instance_material = type('instance_material', (Element,), {'name': 'instance_material'})
instance_material = type('instance_material', (Element,), {'name': 'instance_material'})
instance_node = type('instance_node', (Element,), {'name': 'instance_node'})
instance_physics_material = type('instance_physics_material', (Element,), {'name': 'instance_physics_material'})
instance_physics_model = type('instance_physics_model', (Element,), {'name': 'instance_physics_model'})
instance_physics_scene = type('instance_physics_scene', (Element,), {'name': 'instance_physics_scene'})
instance_rigid_body = type('instance_rigid_body', (Element,), {'name': 'instance_rigid_body'})
instance_rigid_constraint = type('instance_rigid_constraint', (Element,), {'name': 'instance_rigid_constraint'})
instance_visual_scene = type('instance_visual_scene', (Element,), {'name': 'instance_visual_scene'})
int_array = type('int_array', (Element,), {'name': 'int_array'})
interpenetrate = type('interpenetrate', (Element,), {'name': 'interpenetrate'})
jerk = type('jerk', (Element,), {'name': 'jerk'})
axis_info = type('axis_info', (Element,), {'name': 'axis_info'})
joint = type('joint', (Element,), {'name': 'joint'})
joints = type('joints', (Element,), {'name': 'joints'})
keywords = type('keywords', (Element,), {'name': 'keywords'})
kinematics_model = type('kinematics_model', (Element,), {'name': 'kinematics_model'})
kinematics_scene = type('kinematics_scene', (Element,), {'name': 'kinematics_scene'})
kinematics = type('kinematics', (Element,), {'name': 'kinematics'})
lambert = type('lambert', (Element,), {'name': 'lambert'})
latitude = type('latitude', (Element,), {'name': 'latitude'})
layer = type('layer', (Element,), {'name': 'layer'})
library_animation_clips = type('library_animation_clips', (Element,), {'name': 'library_animation_clips'})
library_animations = type('library_animations', (Element,), {'name': 'library_animations'})
library_articulated_systems = type('library_articulated_systems', (Element,), {'name': 'library_articulated_systems'})
library_cameras = type('library_cameras', (Element,), {'name': 'library_cameras'})
library_controllers = type('library_controllers', (Element,), {'name': 'library_controllers'})
library_effects = type('library_effects', (Element,), {'name': 'library_effects'})
library_force_fields = type('library_force_fields', (Element,), {'name': 'library_force_fields'})
library_formulas = type('library_formulas', (Element,), {'name': 'library_formulas'})
library_geometries = type('library_geometries', (Element,), {'name': 'library_geometries'})
library_images = type('library_images', (Element,), {'name': 'library_images'})
library_joints = type('library_joints', (Element,), {'name': 'library_joints'})
library_kinematics_models = type('library_kinematics_models', (Element,), {'name': 'library_kinematics_models'})
library_kinematics_scenes = type('library_kinematics_scenes', (Element,), {'name': 'library_kinematics_scenes'})
library_nodes = type('library_nodes', (Element,), {'name': 'library_nodes'})
library_physics_materials = type('library_physics_materials', (Element,), {'name': 'library_physics_materials'})
library_physics_models = type('library_physics_models', (Element,), {'name': 'library_physics_models'})
library_physics_scenes = type('library_physics_scenes', (Element,), {'name': 'library_physics_scenes'})
library_visual_scenes = type('library_visual_scenes', (Element,), {'name': 'library_visual_scenes'})
light_ambient = type('light_ambient', (Element,), {'name': 'light_ambient'})
light_constant_attenuation = type('light_constant_attenuation', (Element,), {'name': 'light_constant_attenuation'})
light_diffuse = type('light_diffuse', (Element,), {'name': 'light_diffuse'})
light_enable = type('light_enable', (Element,), {'name': 'light_enable'})
light_linear_attenuation = type('light_linear_attenuation', (Element,), {'name': 'light_linear_attenuation'})
light_model_ambient = type('light_model_ambient', (Element,), {'name': 'light_model_ambient'})
light_model_color_control = type('light_model_color_control', (Element,), {'name': 'light_model_color_control'})
light_model_local_viewer_enable = type('light_model_local_viewer_enable', (Element,), {'name': 'light_model_local_viewer_enable'})
light_model_two_side_enable = type('light_model_two_side_enable', (Element,), {'name': 'light_model_two_side_enable'})
light_position = type('light_position', (Element,), {'name': 'light_position'})
light_quadratic_attenuation = type('light_quadratic_attenuation', (Element,), {'name': 'light_quadratic_attenuation'})
light_specular = type('light_specular', (Element,), {'name': 'light_specular'})
light_spot_cutoff = type('light_spot_cutoff', (Element,), {'name': 'light_spot_cutoff'})
light_spot_direction = type('light_spot_direction', (Element,), {'name': 'light_spot_direction'})
light_spot_exponent = type('light_spot_exponent', (Element,), {'name': 'light_spot_exponent'})
lighting_enable = type('lighting_enable', (Element,), {'name': 'lighting_enable'})
lights = type('lights', (Element,), {'name': 'lights'})
limits = type('limits', (Element,), {'name': 'limits'})
axis_info = type('axis_info', (Element,), {'name': 'axis_info'})
prismatic = type('prismatic', (Element,), {'name': 'prismatic'})
revolute = type('revolute', (Element,), {'name': 'revolute'})
line_smooth_enable = type('line_smooth_enable', (Element,), {'name': 'line_smooth_enable'})
line_stipple_enable = type('line_stipple_enable', (Element,), {'name': 'line_stipple_enable'})
line_stipple = type('line_stipple', (Element,), {'name': 'line_stipple'})
line_width = type('line_width', (Element,), {'name': 'line_width'})
line = type('line', (Element,), {'name': 'line'})
linear_attenuation = type('linear_attenuation', (Element,), {'name': 'linear_attenuation'})
spot = type('spot', (Element,), {'name': 'spot'})
linear = type('linear', (Element,), {'name': 'linear'})
lines = type('lines', (Element,), {'name': 'lines'})
linestrips = type('linestrips', (Element,), {'name': 'linestrips'})
link = type('link', (Element,), {'name': 'link'})
linker = type('linker', (Element,), {'name': 'linker'})
locked = type('locked', (Element,), {'name': 'locked'})
logic_op_enable = type('logic_op_enable', (Element,), {'name': 'logic_op_enable'})
logic_op = type('logic_op', (Element,), {'name': 'logic_op'})
longitude = type('longitude', (Element,), {'name': 'longitude'})
lookout = type('lookout', (Element,), {'name': 'lookout'})
magfilter = type('magfilter', (Element,), {'name': 'magfilter'})
mass_frame = type('mass_frame', (Element,), {'name': 'mass_frame'})
rigid_body = type('rigid_body', (Element,), {'name': 'rigid_body'})
mass = type('mass', (Element,), {'name': 'mass'})
shape = type('shape', (Element,), {'name': 'shape'})
material_ambient = type('material_ambient', (Element,), {'name': 'material_ambient'})
material_diffuse = type('material_diffuse', (Element,), {'name': 'material_diffuse'})
material_emission = type('material_emission', (Element,), {'name': 'material_emission'})
material_shininess = type('material_shininess', (Element,), {'name': 'material_shininess'})
material_specular = type('material_specular', (Element,), {'name': 'material_specular'})
material = type('material', (Element,), {'name': 'material'})
matrix = type('matrix', (Element,), {'name': 'matrix'})
max_anisotropy = type('max_anisotropy', (Element,), {'name': 'max_anisotropy'})
mesh = type('mesh', (Element,), {'name': 'mesh'})
minfilter = type('minfilter', (Element,), {'name': 'minfilter'})
mip_bias = type('mip_bias', (Element,), {'name': 'mip_bias'})
mip_max_level = type('mip_max_level', (Element,), {'name': 'mip_max_level'})
mip_min_level = type('mip_min_level', (Element,), {'name': 'mip_min_level'})
mipfilter = type('mipfilter', (Element,), {'name': 'mipfilter'})
mips = type('mips', (Element,), {'name': 'mips'})
create_cube = type('create_cube', (Element,), {'name': 'create_cube'})
create2d = type('create2d', (Element,), {'name': 'create2d'})
create3d = type('create3d', (Element,), {'name': 'create3d'})
model_view_matrix = type('model_view_matrix', (Element,), {'name': 'model_view_matrix'})
modified = type('modified', (Element,), {'name': 'modified'})
modifier = type('modifier', (Element,), {'name': 'modifier'})
morph = type('morph', (Element,), {'name': 'morph'})
motion = type('motion', (Element,), {'name': 'motion'})
multisample_enable = type('multisample_enable', (Element,), {'name': 'multisample_enable'})
Name_array = type('Name_array', (Element,), {'name': 'Name_array'})
emantic = type('emantic', (Element,), {'name': 'emantic'})
newparam = type('newparam', (Element,), {'name': 'newparam'})
ommon = type('ommon', (Element,), {'name': 'ommon'})
node = type('node', (Element,), {'name': 'node'})
normalize_enable = type('normalize_enable', (Element,), {'name': 'normalize_enable'})
nurbs_surface = type('nurbs_surface', (Element,), {'name': 'nurbs_surface'})
nurbs = type('nurbs', (Element,), {'name': 'nurbs'})
optics = type('optics', (Element,), {'name': 'optics'})
orient = type('orient', (Element,), {'name': 'orient'})
origin = type('origin', (Element,), {'name': 'origin'})
orthographic = type('orthographic', (Element,), {'name': 'orthographic'})
p = type('p', (Element,), {'name': 'p'})
edges = type('edges', (Element,), {'name': 'edges'})
faces = type('faces', (Element,), {'name': 'faces'})
lines = type('lines', (Element,), {'name': 'lines'})
pcurves = type('pcurves', (Element,), {'name': 'pcurves'})
ph = type('ph', (Element,), {'name': 'ph'})
solids = type('solids', (Element,), {'name': 'solids'})
triangles = type('triangles', (Element,), {'name': 'triangles'})
trifans = type('trifans', (Element,), {'name': 'trifans'})
tristrips = type('tristrips', (Element,), {'name': 'tristrips'})
wires = type('wires', (Element,), {'name': 'wires'})
parabola = type('parabola', (Element,), {'name': 'parabola'})
param = type('param', (Element,), {'name': 'param'})
ph = type('ph', (Element,), {'name': 'ph'})
phong = type('phong', (Element,), {'name': 'phong'})
physics_material = type('physics_material', (Element,), {'name': 'physics_material'})
physics_model = type('physics_model', (Element,), {'name': 'physics_model'})
physics_scene = type('physics_scene', (Element,), {'name': 'physics_scene'})
plane = type('plane', (Element,), {'name': 'plane'})
point_distance_attenuation = type('point_distance_attenuation', (Element,), {'name': 'point_distance_attenuation'})
point_fade_threshold_size = type('point_fade_threshold_size', (Element,), {'name': 'point_fade_threshold_size'})
point_size_max = type('point_size_max', (Element,), {'name': 'point_size_max'})
point_size_min = type('point_size_min', (Element,), {'name': 'point_size_min'})
point_size = type('point_size', (Element,), {'name': 'point_size'})
point_smooth_enable = type('point_smooth_enable', (Element,), {'name': 'point_smooth_enable'})
polygon_mode = type('polygon_mode', (Element,), {'name': 'polygon_mode'})
polygon_offset_fill_enable = type('polygon_offset_fill_enable', (Element,), {'name': 'polygon_offset_fill_enable'})
polygon_offset_line_enable = type('polygon_offset_line_enable', (Element,), {'name': 'polygon_offset_line_enable'})
polygon_offset_point_enable = type('polygon_offset_point_enable', (Element,), {'name': 'polygon_offset_point_enable'})
polygon_offset = type('polygon_offset', (Element,), {'name': 'polygon_offset'})
polygon_smooth_enable = type('polygon_smooth_enable', (Element,), {'name': 'polygon_smooth_enable'})
polygon_stipple_enable = type('polygon_stipple_enable', (Element,), {'name': 'polygon_stipple_enable'})
polygons = type('polygons', (Element,), {'name': 'polygons'})
polylist = type('polylist', (Element,), {'name': 'polylist'})
prismatic = type('prismatic', (Element,), {'name': 'prismatic'})
profile_BRIDGE = type('profile_BRIDGE', (Element,), {'name': 'profile_BRIDGE'})
profile_CG = type('profile_CG', (Element,), {'name': 'profile_CG'})
profile_COMMON = type('profile_COMMON', (Element,), {'name': 'profile_COMMON'})
verview = type('verview', (Element,), {'name': 'verview'})
profile_GLES = type('profile_GLES', (Element,), {'name': 'profile_GLES'})
profile_GLES2 = type('profile_GLES2', (Element,), {'name': 'profile_GLES2'})
profile_GLSL = type('profile_GLSL', (Element,), {'name': 'profile_GLSL'})
program = type('program', (Element,), {'name': 'program'})
projection_matrix = type('projection_matrix', (Element,), {'name': 'projection_matrix'})
quadratic_attenuation = type('quadratic_attenuation', (Element,), {'name': 'quadratic_attenuation'})
point = type('point', (Element,), {'name': 'point'})
radius = type('radius', (Element,), {'name': 'radius'})
capsule = type('capsule', (Element,), {'name': 'capsule'})
circle = type('circle', (Element,), {'name': 'circle'})
cone = type('cone', (Element,), {'name': 'cone'})
cylinder = type('cylinder', (Element,), {'name': 'cylinder'})
ellipse = type('ellipse', (Element,), {'name': 'ellipse'})
hyperbola = type('hyperbola', (Element,), {'name': 'hyperbola'})
torus = type('torus', (Element,), {'name': 'torus'})
ref_attachment = type('ref_attachment', (Element,), {'name': 'ref_attachment'})
ref = type('ref', (Element,), {'name': 'ref'})
binary = type('binary', (Element,), {'name': 'binary'})
init_from = type('init_from', (Element,), {'name': 'init_from'})
reflective = type('reflective', (Element,), {'name': 'reflective'})
reflectivity = type('reflectivity', (Element,), {'name': 'reflectivity'})
render = type('render', (Element,), {'name': 'render'})
renderable = type('renderable', (Element,), {'name': 'renderable'})
rescale_normal_enable = type('rescale_normal_enable', (Element,), {'name': 'rescale_normal_enable'})
restitution = type('restitution', (Element,), {'name': 'restitution'})
revision = type('revision', (Element,), {'name': 'revision'})
revolute = type('revolute', (Element,), {'name': 'revolute'})
RGB = type('RGB', (Element,), {'name': 'RGB'})
rigid_constraint = type('rigid_constraint', (Element,), {'name': 'rigid_constraint'})
rotate = type('rotate', (Element,), {'name': 'rotate'})
sample_alpha_to_coverage_enable = type('sample_alpha_to_coverage_enable', (Element,), {'name': 'sample_alpha_to_coverage_enable'})
sample_alpha_to_one_enable = type('sample_alpha_to_one_enable', (Element,), {'name': 'sample_alpha_to_one_enable'})
sample_coverage_enable = type('sample_coverage_enable', (Element,), {'name': 'sample_coverage_enable'})
sample_coverage = type('sample_coverage', (Element,), {'name': 'sample_coverage'})
sampler_image = type('sampler_image', (Element,), {'name': 'sampler_image'})
sampler_states = type('sampler_states', (Element,), {'name': 'sampler_states'})
sampler = type('sampler', (Element,), {'name': 'sampler'})
nterpolation = type('nterpolation', (Element,), {'name': 'nterpolation'})
sampler1D = type('sampler1D', (Element,), {'name': 'sampler1D'})
sampler2D = type('sampler2D', (Element,), {'name': 'sampler2D'})
sampler3D = type('sampler3D', (Element,), {'name': 'sampler3D'})
samplerCUBE = type('samplerCUBE', (Element,), {'name': 'samplerCUBE'})
samplerDEPTH = type('samplerDEPTH', (Element,), {'name': 'samplerDEPTH'})
samplerRECT = type('samplerRECT', (Element,), {'name': 'samplerRECT'})
scale = type('scale', (Element,), {'name': 'scale'})
scene = type('scene', (Element,), {'name': 'scene'})
scissor_test_enable = type('scissor_test_enable', (Element,), {'name': 'scissor_test_enable'})
scissor = type('scissor', (Element,), {'name': 'scissor'})
semantic = type('semantic', (Element,), {'name': 'semantic'})
setparam = type('setparam', (Element,), {'name': 'setparam'})
shade_model = type('shade_model', (Element,), {'name': 'shade_model'})
shader = type('shader', (Element,), {'name': 'shader'})
shape = type('shape', (Element,), {'name': 'shape'})
shells = type('shells', (Element,), {'name': 'shells'})
shininess = type('shininess', (Element,), {'name': 'shininess'})
SIDREF_array = type('SIDREF_array', (Element,), {'name': 'SIDREF_array'})
size_exact = type('size_exact', (Element,), {'name': 'size_exact'})
size_ratio = type('size_ratio', (Element,), {'name': 'size_ratio'})
size = type('size', (Element,), {'name': 'size'})
create_cube = type('create_cube', (Element,), {'name': 'create_cube'})
create3d = type('create3d', (Element,), {'name': 'create3d'})
skeleton = type('skeleton', (Element,), {'name': 'skeleton'})
skew = type('skew', (Element,), {'name': 'skew'})
skin = type('skin', (Element,), {'name': 'skin'})
source_data = type('source_data', (Element,), {'name': 'source_data'})
source = type('source', (Element,), {'name': 'source'})
sources = type('sources', (Element,), {'name': 'sources'})
specular = type('specular', (Element,), {'name': 'specular'})
speed = type('speed', (Element,), {'name': 'speed'})
sphere = type('sphere', (Element,), {'name': 'sphere'})
spline = type('spline', (Element,), {'name': 'spline'})
nterpolation = type('nterpolation', (Element,), {'name': 'nterpolation'})
spot = type('spot', (Element,), {'name': 'spot'})
spring = type('spring', (Element,), {'name': 'spring'})
states = type('states', (Element,), {'name': 'states'})
static_friction = type('static_friction', (Element,), {'name': 'static_friction'})
stencil_clear = type('stencil_clear', (Element,), {'name': 'stencil_clear'})
stencil_func_separate = type('stencil_func_separate', (Element,), {'name': 'stencil_func_separate'})
stencil_func = type('stencil_func', (Element,), {'name': 'stencil_func'})
stencil_mask_separate = type('stencil_mask_separate', (Element,), {'name': 'stencil_mask_separate'})
stencil_mask = type('stencil_mask', (Element,), {'name': 'stencil_mask'})
stencil_op_separate = type('stencil_op_separate', (Element,), {'name': 'stencil_op_separate'})
stencil_op = type('stencil_op', (Element,), {'name': 'stencil_op'})
stencil_target = type('stencil_target', (Element,), {'name': 'stencil_target'})
stencil_test_enable = type('stencil_test_enable', (Element,), {'name': 'stencil_test_enable'})
stiffness = type('stiffness', (Element,), {'name': 'stiffness'})
subject = type('subject', (Element,), {'name': 'subject'})
surface_curves = type('surface_curves', (Element,), {'name': 'surface_curves'})
surface = type('surface', (Element,), {'name': 'surface'})
surfaces = type('surfaces', (Element,), {'name': 'surfaces'})
swept_surface = type('swept_surface', (Element,), {'name': 'swept_surface'})
swing_cone_and_twist = type('swing_cone_and_twist', (Element,), {'name': 'swing_cone_and_twist'})
target_value = type('target_value', (Element,), {'name': 'target_value'})
target = type('target', (Element,), {'name': 'target'})
targets = type('targets', (Element,), {'name': 'targets'})
bind_material = type('bind_material', (Element,), {'name': 'bind_material'})
formula = type('formula', (Element,), {'name': 'formula'})
instance_rigid_body = type('instance_rigid_body', (Element,), {'name': 'instance_rigid_body'})
kinematics_model = type('kinematics_model', (Element,), {'name': 'kinematics_model'})
kinematics = type('kinematics', (Element,), {'name': 'kinematics'})
light = type('light', (Element,), {'name': 'light'})
motion = type('motion', (Element,), {'name': 'motion'})
optics = type('optics', (Element,), {'name': 'optics'})
rigid_body = type('rigid_body', (Element,), {'name': 'rigid_body'})
verview = type('verview', (Element,), {'name': 'verview'})
technique_hint = type('technique_hint', (Element,), {'name': 'technique_hint'})
technique_override = type('technique_override', (Element,), {'name': 'technique_override'})
texcombiner = type('texcombiner', (Element,), {'name': 'texcombiner'})
texcoord = type('texcoord', (Element,), {'name': 'texcoord'})
texenv = type('texenv', (Element,), {'name': 'texenv'})
texture_env_color = type('texture_env_color', (Element,), {'name': 'texture_env_color'})
texture_env_mode = type('texture_env_mode', (Element,), {'name': 'texture_env_mode'})
texture_pipeline = type('texture_pipeline', (Element,), {'name': 'texture_pipeline'})
texture = type('texture', (Element,), {'name': 'texture'})
texture1D_enable = type('texture1D_enable', (Element,), {'name': 'texture1D_enable'})
texture1D = type('texture1D', (Element,), {'name': 'texture1D'})
texture2D_enable = type('texture2D_enable', (Element,), {'name': 'texture2D_enable'})
texture2D = type('texture2D', (Element,), {'name': 'texture2D'})
texture3D_enable = type('texture3D_enable', (Element,), {'name': 'texture3D_enable'})
texture3D = type('texture3D', (Element,), {'name': 'texture3D'})
textureCUBE_enable = type('textureCUBE_enable', (Element,), {'name': 'textureCUBE_enable'})
textureCUBE = type('textureCUBE', (Element,), {'name': 'textureCUBE'})
textureDEPTH_enable = type('textureDEPTH_enable', (Element,), {'name': 'textureDEPTH_enable'})
textureDEPTH = type('textureDEPTH', (Element,), {'name': 'textureDEPTH'})
textureRECT_enable = type('textureRECT_enable', (Element,), {'name': 'textureRECT_enable'})
textureRECT = type('textureRECT', (Element,), {'name': 'textureRECT'})
time_step = type('time_step', (Element,), {'name': 'time_step'})
title = type('title', (Element,), {'name': 'title'})
translate = type('translate', (Element,), {'name': 'translate'})
transparency = type('transparency', (Element,), {'name': 'transparency'})
transparent = type('transparent', (Element,), {'name': 'transparent'})
unit = type('unit', (Element,), {'name': 'unit'})
unnormalized = type('unnormalized', (Element,), {'name': 'unnormalized'})
usertype = type('usertype', (Element,), {'name': 'usertype'})
v = type('v', (Element,), {'name': 'v'})
value = type('value', (Element,), {'name': 'value'})
velocity = type('velocity', (Element,), {'name': 'velocity'})
vertex_weights = type('vertex_weights', (Element,), {'name': 'vertex_weights'})
wrap_p = type('wrap_p', (Element,), {'name': 'wrap_p'})
wrap_s = type('wrap_s', (Element,), {'name': 'wrap_s'})
wrap_t = type('wrap_t', (Element,), {'name': 'wrap_t'})
xfov = type('xfov', (Element,), {'name': 'xfov'})
xmag = type('xmag', (Element,), {'name': 'xmag'})
yfov = type('yfov', (Element,), {'name': 'yfov'})
ymag = type('ymag', (Element,), {'name': 'ymag'})
zfar = type('zfar', (Element,), {'name': 'zfar'})
znear = type('znear', (Element,), {'name': 'znear'})

# builtins require an underscore.
_input = type('input', (closed_tag,), {'name': 'input'})
_float = type('float', (Element,), {'name': 'float'})
_pass = type('pass', (Element,), {'name': 'pass'})
_max = type('max', (Element,), {'name': 'max'})
_min = type('min', (Element,), {'name': 'min'})
_import = type('import', (Element,), {'name': 'import'})
_hex = type('hex', (Element,), {'name': 'hex'})
_format = type('format', (Element,), {'name': 'format'})


class xml_header():
    """xml

    Returns:
        str: <?xml version="1.0"?>
    """
    def __str__(self):
        return '<?xml version="1.0"?>'


'''

# The asset tag describes the author and environment of the file
TMP_ASSET = asset(
    author("alorino"),
    up_axis("Y_UP")
)

# common cameras are perspective or orthographic
LIB_CAMERAS = library_cameras(
    camera(_id="PerspCamera", _name="PerspCamera").appendChild(
        optics(
            technique_common(
                perspective(
                    yfov(37.8493),
                    aspect_ratio(1),
                    znear(10),
                    zfar(1000)
                )
            )
        )
    )
)

# Common lights are point, spot or directional
LIB_LIGHTS = library_lights(
    light(_id="pointLightShape1-lib", _name="pointLightShape1").appendChild(
        technique_common(
            point(
                color("1 1 1"),  # need to figure this out
                constant_attenuation(1),
                linear_attenuation(0),
                quadratic_attenuation(0)
            )
        )
    )
)

# Materials instance effects
LIB_MAT = library_materials(
    material(_id="Blue", _name="Blue").appendChild(
        instance_effect(_url="#Blue-fx")
    )
)

# Common effects act like the OpenGL 1 state
LIB_FX = library_effects(
    effect(_id="Blue-fx").appendChild(
        profile_COMMON(
            technique(_sid="common").appendChild(
                phong(
                    emission(
                        color("0 0 0 1"),
                        index_of_refraction(
                            _float(0)
                        )
                    )
                )
            )
        )
    )
)

# Geometry describes the OpenGL attributes
LIB_GEOM = library_geometries(
    geometry(_id="box-lib", _name="box").appendChild(
        mesh(
            source(_id="box-lib-positions", _name="position"),
            source(_id="box-lib-normals", _name="normal"),
            vertices(_id="box-lib-vertices").appendChild(
                _input(_semantic="POSITION", _source="#box-lib-positions")
            ),
            polylist(_count="6", _material="BlueSG").appendChild(
                _input(_offset="0", _semantic="VERTEX", _source="#box-lib-vertices"),
                _input(_offset="1", _semantic="NORMAL", _source="#box-lib-normals"),
                vcount("4 4 4 4 4 4"),
                p("0 0 2 1 3 2 1 3 0 4 1 5 5 6 4 7")
            )
        )
    )
)

# The scene tag specifies a visual scene and sometimes a phyiscs scene
LIB_SCENES = library_visual_scenes(
    visual_scene(_id="VisualSceneNode", _name="untitled").appendChild(
        node(_id="Camera", _name="Camera")
    )
)

TMP_SCENE = scene(
    instance_visual_scene(_url="#VisualSceneNode")
)

# Extras are a way of extending the format
TMP_EXTRA = extra(
    technique(_profile="steveT").appendChild(
        test_element(_id="my_test_element", _attr1="value1", _attr2="value2").appendChild(
            "this is some text"
        )
    )
)

'''

"""

library_cameras(
    camera(_id="Camera-camera", _name="Camera").html(
        optics(
            technique_common(
                perspective(
                    xfov("49.13434", _sid="xfov")
                    aspect_ratio("1.777778")  # NOTE - underscored attributes?. - likely to break?
                    znear("0.1", _sid="znear")
                    zfar("100", _sid="zfar")
                )
            )
        )
    )
)

source(_id="Cube-mesh-normals").html(
    float_array(_id="Cube-mesh-normals-array", _count="18").html("0 0 -1 0 0 1 1 -2.83e-7 0 -2.83e-7 -1 0 -1 2.23e-7 -1.34e-7 2.38e-7 1 2.08e-7"),
        technique_common(
            accessor(_source="#Cube-mesh-normals-array", _count="6", _stride="3").html(
                param(_name="X" type="float"),
                param(_name="Y" type="float"),
            param(_name="Z" type="float")
        )
    )
)

vertices(_id="Cube-mesh-vertices").html(
    input(_semantic="POSITION", _source="#Cube-mesh-positions")
)


COLLADA(
    asset(
        created("2005-06-27T21:00:00Z"),
        keywords("COLLADA inthange"),
        created("2005-06-27T21:00:00Z"),
        unit(_name="nautical_league", _meter="5556.0"),
        up_axis("Z_UP")
    )
)

light(_id="blue").html(
    technique_common(
        directional(
        color(0.1 0.1 0.5)
        )
    )
)


camera(_name="eyepoint").html(
    optics(
        technique_common("....")
        technique profile="MyFancyGIRenderer">
            param("180.0", _name="FocalLength", _type="float")
            param("5.6", _name="Aperture", _type="float")
        )
    ),
    imager(
        technique(_profile="MyFancyGIRenderer").html(
            param("200.0", _name="ShutterSpeed", _type="float")
            param("0.2", _name="RedGain", _type="float")
            param("0.22", _name="GreenGain", _type="float")
            param("0.25", _name="BlueGain", _type="float")
            param("2.2", _name="RedGamma", _type="float")
            param("2.1", _name="GreenGamma", _type="float")
            param("2.17", _name="BlueGamma", _type="float")
            param("0.17", _name="BloomPixelLeak", _type="float")
            param("InvSquare", _name="BloomFalloff", _type="Name")
        )
    )
)


+ asset
+ + contributor
+ + + author
+ + + authoring_tool
+ + + created
+ + + unit
+ + + up_axi
+ library_cameras
+ + camera
+ + + optics
+ + + + technique_common
+ + + + + perspective
+ + + + + + xfov
+ + + + + + aspect_ratio
+ + + + + + znear
+ + + + + + zfar
+ library_lights
+ + light
+ + + technique_common
+ + + + point
+ + + + + color
+ + + + + constant_attenuation
+ + + + + linear_attenuation
+ + + + + quadratic_attenuation
+ + + extra
+ + + + technique
............
+ library_visual_scenes
+ + visual_scene
+ + + node
+ + + + translate
+ + + + rotate
+ + + + rotate
+ + + + rotate
+ + + + scale
+ + + + instance_geometry
+ + + + + bind_material
+ + + + + + technique_common
+ + + + + + + instance_material
+ + + node
+ + + + translate
+ + + + rotate
+ + + + rotate
+ + + + rotate
+ + + + scale
+ + + + instance_camera
+ scene
+ + instance_visual_scene

"""



"""
    domonic.webapi.sse
    ====================================
    https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events

    TODO - consider a blocking verison?
    TODO - port one of the polyfills?
    https://github.com/EventSource/eventsource/blob/master/lib/eventsource.js
    https://github.com/remy/polyfills/blob/master/EventSource.js
    https://github.com/Yaffle/EventSource/blob/master/src/eventsource.js

    # TESTING - useage example... clone this and point at the stream
    # https://github.com/byteface/SSELoggerDemo

"""

from domonic.events import MessageEvent, EventTarget


class EventSource(EventTarget):
    # https://developer.mozilla.org/en-US/docs/Web/API/EventSource

    CONNECTING = 0
    OPEN = 1
    CLOSED = 2

    def __init__(self, url, eventSourceInitDict=None, **kwargs):
        """
        The EventSource interface provides the ability to asynchronously receive messages from a server using simple push paradigm.
        """
        eventSourceInitDict = eventSourceInitDict or kwargs
        super().__init__()
        self._url = url
        self._readyState = EventSource.CLOSED
        self._withCredentials = eventSourceInitDict.get('withCredentials', False)
        self._lastEventId = eventSourceInitDict.get('lastEventId', None)

        self._onmessage = eventSourceInitDict.get('onmessage', None)
        self._onerror = eventSourceInitDict.get('onerror', None)
        self._onopen = eventSourceInitDict.get('onopen', None)
        self._onreadystatechange = eventSourceInitDict.get('onreadystatechange', None)

        # from sseclient import SSEClient
        # messages = SSEClient(url)
        # self._readyState = EventSource.OPEN
        # for msg in messages:
        #     self.onmessage(MessageEvent(msg))
        #     print(msg)
        # self._readyState = EventSource.CLOSED

        def _do_stuff():
            from sseclient import SSEClient
            messages = SSEClient(url)
            self._readyState = EventSource.OPEN
            # self.onopen(MessageEvent(messages))
            for msg in messages:
                self.onmessage(MessageEvent(msg))
                # print(msg)
            self._readyState = EventSource.CLOSED

        # TODO - use a thread or it will block the main thread
        import threading

        # def _on_message(msg):
        #     self.onmessage(MessageEvent(msg))
        #     print(msg)
        #     self._readyState = EventSource.OPEN

        # def _on_error(err):
        #     self.onerror(err)
        #     self._readyState = EventSource.CLOSED

        # def _on_open(msg):
        #     self.onopen(msg)
        #     self._readyState = EventSource.OPEN

        # def _on_ready_state_change(msg):
        #     self.onreadystatechange(msg)
        #     self._readyState = EventSource.OPEN

        # def _on_close(msg):
        #     self.onclose(msg)
        #     self._readyState = EventSource.CLOSED

        self._thread = threading.Thread(target=_do_stuff)
        self._thread.start()

        # print('this is so cool!')

    @property
    def readyState(self):
        """ A number representing the state of the connection.
        Possible values are CONNECTING (0), OPEN (1), or CLOSED (2). """
        return self._readyState

    @property
    def url(self):
        """ A DOMString representing the URL of the source. """
        return self._url

    @property
    def withCredentials(self):
        """ A boolean value indicating whether the EventSource object was
          instantiated with cross-origin (CORS) credentials
        set (true), or not (false, the default). """
        return self._withCredentials

    def close(self):
        """ Closes the connection to the EventSource. """
        self._readyState = EventSource.CLOSED
        # close the thread
        self._thread.join()
        self._thread = None

    def onreadystatechange(self, event):
        """ Called when the state of the connection changes. """
        pass

    def onmessage(self, event):
        """ Called when a message is received. """
        if self._onmessage is not None:
            self._onmessage(event)

    def onerror(self, event):
        """ Called when an error occurs. """
        pass

    def onopen(self, event):
        """ Called when the connection is established. """
        pass

    # '''
    # NOTE - if we write custom one rather than use lib...
    # def _poll(self):

    #     if self._readyState == EventSource.CLOSED:
    #         return

    #     url = 'http://domain.com/events'
    #     headers = {
    #         'Accept': 'text/event-stream',
    #         'Cache-Control': 'no-cache',
    #         'X-requested-with': 'XMLHttpRequest'
    #     }
    #     if self._lastEventId != None:
    #         headers['Last-Event-ID'] = self._lastEventId

    #     import requests
    #     response = requests.get(url, stream=True, headers=headers)
    #    # TODO - parse repsonse
    # '''
